"""
Peptide record (PEPREC).

TODO: Move module to ms2pip
"""

from typing import List, Optional

import pandas as pd


class InvalidPeprecError(Exception):
    """Invalid PEPREC file."""

    pass


class PeptideRecord:
    """Peptide record (PEPREC)."""

    def __init__(
        self,
        path: Optional[str] = None,
        context: str = "default",
        extra_required_columns: Optional[List[str]] = None,
    ):
        """
        Peptide record (PEPREC).

        Parameters
        ----------
        path: str, None
            Path to PEPREC file. If not None, file is read on instance creation.
        context: str
            Context of PEPREC. Is used for determining the required columns. Can be any
            of the following: {default, ms2rescore}.
        extra_required_columns: List[str]
            Extra required columns to be validated.

        Attributes
        ----------
        df: pandas.DataFrame
            DataFrame containing peptide record content.

        Methods
        -------
        from_csv(path, **kwargs)
            Read PEPREC from CSV.
        to_csv(path, **kwargs)
            Save PEPREC to CSV, if Path is None, overwrite existing PEPREC file.

        """
        self.path = path
        self.context = context
        self.extra_required_columns = extra_required_columns
        self.df = None

        if self.path:
            self.read_csv(self.path)

    def __repr__(self):
        """Get string representation of PeptideRecord."""
        return self.df.__repr__()

    @property
    def df(self) -> Optional[pd.DataFrame]:
        """Get DataFrame with PeptideRecord."""
        return self._df

    @df.setter
    def df(self, value: Optional[pd.DataFrame]):
        """Set DataFrame with PeptideRecord."""
        if isinstance(value, pd.DataFrame):
            self._validate_column_names(value)
        self._df = value

    def read_csv(self, path: str, **kwargs):
        """Read PEPREC from CSV."""
        self._validate_header(path)
        sep = self._infer_separator(path)
        self.df = pd.read_csv(path, sep=sep, index_col=None, **kwargs)
        self.df["modifications"] = self.df["modifications"].fillna("-")

    def to_csv(self, path: Optional[str] = None, **kwargs):
        """Save PEPREC to CSV, if Path is None, overwrite existing PEPREC file."""
        if not path and self.path:
            path = self.path
        elif not path and not self.path:
            raise ValueError("No known path to write PEPREC file.")
        self.df.to_csv(path, sep=" ", index=False, float_format="%.3f", **kwargs)

    def reorder_columns(self):
        """Reorder columns with canonical PeptideRecord columns first."""
        original_columns = self.df.columns.to_list()
        for col in self._required_columns:
            if col in original_columns:
                original_columns.remove(col)
            else:
                self._required_columns.remove(col)
        reordered_columns = self._required_columns + original_columns
        self.df = self.df[reordered_columns]

    @property
    def _required_columns(self):
        """Get required columns depending on PeptideRecord context."""
        required_columns = [
            "spec_id",
            "peptide",
            "modifications",
            "charge",
        ]
        if self.context == "ms2rescore":
            required_columns.extend(
                ["psm_score", "observed_retention_time",]
            )
        elif self.context == "retention_time":
            required_columns.remove("modifications")
        return required_columns

    def _validate_header(self, path: str):
        """Validate PEPREC header."""
        with open(path, "rt") as f:
            line = f.readline()
            if line[:7] != "spec_id":
                raise InvalidPeprecError("PEPREC header should start with `spec_id`.")

    def _validate_column_names(self, df: pd.DataFrame):
        """Validate header of PEPREC file."""
        for col in self._required_columns:
            if col not in df.columns:
                raise InvalidPeprecError(
                    f"Required column `{col}` missing from header."
                )

    @classmethod
    def from_dataframe(cls, dataframe: pd.DataFrame):
        """Create PeptideRecord instance with data from a pandas DataFrame."""
        peprec = cls()
        peprec.df = dataframe
        peprec.reorder_columns()
        return peprec

    @classmethod
    def from_csv(cls, *args, **kwargs):
        """Create PeptideRecord instance with data from CSV file."""
        peprec = cls()
        peprec.read_csv(*args, **kwargs)
        return peprec

    @staticmethod
    def _infer_separator(path: str) -> str:
        """Infer separator in PEPREC file."""
        with open(path, "rt") as f:
            line = f.readline()
            separator = line[7]
        return separator

    def validate_decoy_presence(self):
        "check whether decoys are present in peprec file"

        if sum(self.df["Label"] == -1) == 0:
            raise InvalidPeprecError("No decoys present, make sure to provide decoys along targets")
        else:
            pass



"""Extract PeptideRecord and search engine features from identification file."""

import logging
import os
import re
from abc import ABC, abstractmethod
from typing import Dict, Tuple, Union, List

import numpy as np
import pandas as pd
from pyteomics import tandem, mzid
from tqdm import tqdm

from ms2rescore._exceptions import MS2RescoreError
from ms2rescore.maxquant import MSMSAccessor
from ms2rescore.parse_mgf import parse_mgf
from ms2rescore.peptide_record import PeptideRecord
from ms2rescore.peptideshaker import ExtendedPsmReportAccessor
from ms2rescore.percolator import PercolatorIn, run_percolator_converter


logger = logging.getLogger(__name__)


class IDFileParserError(MS2RescoreError):
    """Error parsing ID file."""

    pass


def parse_mgf_title_rt(
    path_to_mgf: Union[str, os.PathLike]
) -> Tuple[Dict[int, str], Dict[int, float]]:
    """Parse MGF file to extract title and retention time fields, by spectrum index."""
    titles = dict()
    retention_times = dict()
    with open(path_to_mgf, "rt") as mgf_in:
        index = 0
        for line in mgf_in:
            if line[0] == "B":
                if line.strip() == "BEGIN IONS":
                    index += 1
            if line[0] == "T":
                if line.startswith("TITLE="):
                    titles[index] = line[6:].strip()
            if line[0] == "R":
                if line.startswith("RTINSECONDS="):
                    retention_times[index] = float(line[12:].strip())
    return titles, retention_times


class _Pipeline(ABC):
    """ABC for pipeline from ID file to PeptideRecord and search engine features."""

    def __init__(self, config: Dict, output_basename: Union[str, os.PathLike]) -> None:
        # Attributes from config and output_basename
        self.path_to_id_file = config["general"]["identification_file"]
        self.passed_mgf_path = config["general"]["mgf_path"]
        self.output_basename = output_basename
        self.modification_list = config["ms2pip"]["modifications"]
        self.id_decoy_pattern = config["general"]["id_decoy_pattern"]
        self.log_level = config["general"]["log_level"]

        # General private attributes
        self._pin_spec_id_patterns = {
            "generic": r".+_([0-9]+)_[0-9]+_[0-9]+",
            "tandem": r".+_([0-9]+)_[0-9]+_[0-9]+",
            "msgfplus": r".+_SII_([0-9]+)_[0-9]+_[0-9]+_[0-9]+",
            # "USI": r"mzspec:PXD[0-9]{6}:[^\s\:]*:scan:([0-9]+)"
        }

        # Private attributes specific to pipeline, override these in each subclass
        self._path_to_original_pin = None
        self._pin_converter_name = None
        self._pin_spec_id_style = "generic"
        self._pin_modification_style = "infer"

        # Private attributes used by methods
        self._original_pin = None

    @staticmethod
    def _validate_mgf_path(
        passed_path: Union[None, str, os.PathLike],
        default_dir: Union[str, os.PathLike],
        expected_rootname: str,
        expected_ext: str = ".mgf",
    ) -> Union[str, os.PathLike]:
        """
        Infer MGF path from passed path and expected filename (e.g. from ID file).

        Parameters
        ----------
        passed_path : None, string, os.PathLike
            user-defined path to MGF file or directory containing MGF file
        default_dir : str, os.PathLike
            default directory for MGF file, used in combination with `expected_rootname`
            and `expected_ext` if `passed_path` is None
        expected_rootname : str, os.PathLike
            rootname of MGF file, as expected from, e.g., identification file
        expected_ext : str, optional
            expected filename extension, including period, default: ".mgf"

        """
        # Make sure that expected_rootname is in fact rootname without expected ext
        expected_rootname = os.path.basename(expected_rootname)
        if expected_rootname.endswith(expected_ext):
            expected_rootname = re.sub(".mgf$", "", expected_rootname)

        # If no mgf path configured, return expected rootname + ext in default directory
        if not passed_path:
            path_to_mgf_file = os.path.join(
                default_dir, expected_rootname + expected_ext
            )

        # If passed path is directory, join with expected rootname + ext
        elif os.path.isdir(passed_path):
            path_to_mgf_file = os.path.join(
                passed_path, expected_rootname + expected_ext
            )

        # If passed path is file, use that, but warn if basename doesn't match expected
        elif os.path.isfile(passed_path):
            passed_rootname = os.path.splitext(os.path.basename(passed_path))[0]
            if passed_rootname != expected_rootname:
                logger.debug(
                    "Passed MGF name root `%s` does not match MGF name root `%s` from "
                    "identifications file. Continuing with passed MGF name.",
                    passed_rootname,
                    expected_rootname,
                )
            path_to_mgf_file = passed_path

        else:
            raise IDFileParserError(
                "Configured `mgf_path` must be None or a path to an existing file or "
                "directory."
            )

        return path_to_mgf_file

    def _run_percolator_converter(self):
        """Run Percolator converter with settings specific to pipeline."""
        run_percolator_converter(
            self._pin_converter_name,
            self.path_to_id_file,
            self._path_to_original_pin,
            id_decoy_pattern=self.id_decoy_pattern,
            log_level=self.log_level,
        )

    @property
    def path_to_mgf_file(self) -> Union[str, os.PathLike]:
        """Get path to MGF file, inferred from mgf_path (which can also be to a dir)."""
        path_to_mgf_file = self._validate_mgf_path(
            self.passed_mgf_path,
            os.path.dirname(self.path_to_id_file),
            os.path.basename(os.path.splitext(self.path_to_id_file)[0]),
        )
        return path_to_mgf_file

    @property
    def original_pin(self) -> PercolatorIn:
        """Get PercolatorIn object from identification file."""
        if not self._original_pin:
            if not self._path_to_original_pin:
                raise TypeError("To load PIN, path_to_pin cannot be None.")
            elif not os.path.isfile(self._path_to_original_pin):
                self._run_percolator_converter()
            self._original_pin = PercolatorIn(self._path_to_original_pin)
            self._original_pin.modification_mapping_from_list(
                self.modification_list, label_style=self._pin_modification_style
            )
        return self._original_pin

    def peprec_from_pin(self) -> PeptideRecord:
        """Get PeptideRecord from PIN file and MGF file."""
        # Get peprec
        peprec = self.original_pin.to_peptide_record(
            spectrum_index_pattern=self._pin_spec_id_patterns[self._pin_spec_id_style]
        )
        titles, retention_times = parse_mgf_title_rt(self.path_to_mgf_file)
        peprec.df["spec_id"] = peprec.df["spec_id"].map(titles)
        if "observed_retention_time" not in peprec.df.columns:
            # Map MGF titles and observed retention times
            peprec.df["observed_retention_time"] = peprec.df["spec_id"].map(
                retention_times
            )

        if not ~peprec.df["observed_retention_time"].isna().any():
            raise IDFileParserError(
                "Could not map all MGF retention times to spectrum indices."
            )
        if not ~peprec.df["spec_id"].isna().any():
            raise IDFileParserError("Could not map all MGF titles to spectrum indices.")

        return peprec

    @abstractmethod
    def get_peprec(self) -> PeptideRecord:
        """Get PeptideRecord."""
        raise NotImplementedError

    @abstractmethod
    def get_search_engine_features(self) -> pd.DataFrame:
        """Get pandas.DataFrame with search engine features."""
        raise NotImplementedError


class PinPipeline(_Pipeline):
    """Percolator IN file to PeptideRecord and search engine features."""

    def __init__(self, config: Dict, output_basename: Union[str, os.PathLike]) -> None:
        super().__init__(config, output_basename)

        # Private attributes specific to pipeline
        self._path_to_original_pin = self.path_to_id_file
        self._pin_converter_name = None
        self._pin_spec_id_style = "generic"
        self._pin_modification_style = "infer"

    def get_peprec(self) -> PeptideRecord:
        """Get PeptideRecord from original PIN file."""
        return self.peprec_from_pin()

    def get_search_engine_features(self) -> pd.DataFrame:
        """Get pandas.DataFrame with search engine features."""
        return self.original_pin.get_feature_table()


class MSGFPipeline(_Pipeline):
    """MSGFPlus mzid to PeptideRecord and search engine features."""

    def __init__(self, config: Dict, output_basename: Union[str, os.PathLike]) -> None:
        super().__init__(config, output_basename)

        # Private attributes specific to pipeline
        self._path_to_original_pin = output_basename + "_original.pin"
        self._pin_converter_name = "msgf2pin"
        self._pin_spec_id_style = "msgfplus"
        self._pin_modification_style = "infer"

    def get_peprec(self) -> PeptideRecord:
        """Get PeptideRecord from original PIN file."""
        return self.peprec_from_pin()

    def get_search_engine_features(self) -> pd.DataFrame:
        """Get pandas.DataFrame with search engine features."""
        return self.original_pin.get_feature_table()


class TandemPipeline(_Pipeline):
    """X!Tandem XML to PeptideRecord and search engine features."""

    def __init__(self, config: Dict, output_basename: Union[str, os.PathLike]) -> None:
        super().__init__(config, output_basename)

        # Private attributes specific to pipeline
        self._path_to_original_pin = output_basename + "_original.pin"
        self._pin_converter_name = "tandem2pin"
        self._pin_spec_id_style = "tandem"
        self._pin_modification_style = "mass_shift"

    # Override method to use spectrum filenames found in PIN spec_ids as expected_root
    @property
    def path_to_mgf_file(self) -> Union[str, os.PathLike]:
        """Get path to MGF file, inferred from mgf_path (which can also be to a dir)."""
        path_to_mgf_file = self._validate_mgf_path(
            self.passed_mgf_path,
            os.path.dirname(self.path_to_id_file),
            self.original_pin.get_spectrum_filename(),
        )
        return path_to_mgf_file

    def get_peprec(self) -> PeptideRecord:
        """Convert X!Tandem XML file and PIN to PEPREC."""
        # Load tandem dataframe with Pyteomics
        logger.debug("Converting X!Tandem XML to PEPREC...")
        tandem_df = tandem.DataFrame(self.path_to_id_file)
        tandem_df["id"] = tandem_df["id"].astype(int)
        if "RTINSECONDS" in tandem_df["scan"].loc[0]:
            # Sometimes "RTINSECONDS" is in scan number column...
            tandem_df["scan"] = tandem_df["scan"].str.replace(" RTINSECONDS.*", "")

        tandem_peprec_mapping = {
            "scan": "spec_id",
            "seq": "peptide",
            "z": "charge",
            "rt": "observed_retention_time",
            "expect": "psm_score",
            "hyperscore": "hyperscore_tandem",
            "id": "tandem_id",
        }

        peprec_df = tandem_df[tandem_peprec_mapping.keys()].rename(
            columns=tandem_peprec_mapping
        )
        # Set PSM score as -log(e-value)
        peprec_df["psm_score"] = -np.log(peprec_df["psm_score"])

        logger.debug("Adding modifications from original PIN to PEPREC...")
        pin = self.original_pin
        pin.add_peprec_modifications_column()
        pin.add_spectrum_index_column(label="tandem_id")
        pin.df["ModPeptide"] = pin.df["Peptide"]
        peprec_df = peprec_df.merge(
            pin.df[
                [
                    "modifications",
                    "tandem_id",
                    "hyperscore",
                    "Label",
                    "ModPeptide",
                    "Proteins",
                ]
            ],
            on="tandem_id",
        )
        # Validate merge by comparing the hyperscore columns
        if not (peprec_df["hyperscore_tandem"] == peprec_df["hyperscore"]).all():
            raise IDFileParserError(
                "Could not merge tandem xml and generated pin files."
            )
        peprec_df.drop(columns=["tandem_id", "hyperscore_tandem"], inplace=True)

        peprec = PeptideRecord.from_dataframe(peprec_df)
        peprec.reorder_columns()
        return peprec

    def get_search_engine_features(self) -> pd.DataFrame:
        """Get pandas.DataFrame with search engine features."""
        return self.original_pin.get_feature_table()


class MaxQuantPipeline(_Pipeline):
    """MaxQuant msms.txt to PeptideRecord and search engine features."""

    def __init__(self, config: Dict, output_basename: Union[str, os.PathLike]) -> None:
        super().__init__(config, output_basename)

        # Private attributes, specific to this pipeline
        mq_conf = config["maxquant_to_rescore"]
        self._modification_mapping = mq_conf["modification_mapping"]
        self._fixed_modifications = mq_conf["fixed_modifications"]
        self._path_to_new_mgf = None
        self._msms_df = None

    @property
    def path_to_mgf_file(self):
        """Get path to single unified MGF file."""
        if not self._path_to_new_mgf:
            logger.warning(
                "`_path_to_new_mgf` is not set yet; first run the `parse_mgf_files` "
                "method"
            )
        else:
            return self._path_to_new_mgf

    @property
    def original_pin(self):
        """Get PercolatorIn object from identification file."""
        raise NotImplementedError(
            "Property `original_pin` is not implemented in class `MaxQuantPipeline`."
        )

    def peprec_from_pin(self):
        """Get PeptideRecord from PIN file and MGF file."""
        raise NotImplementedError(
            "Method `peprec_from_pin` is not implemented in class `MaxQuantPipeline`."
        )

    @property
    def msms_df(self):
        """Get msms.txt identification results."""
        if self._msms_df is None:
            self._msms_df = pd.DataFrame.msms.from_file(self.path_to_id_file)
        return self._msms_df

    def parse_mgf_files(self, peprec):
        """Parse multiple MGF files into one for MS²PIP."""
        logger.debug("Parsing MGF files into one for MS²PIP")
        path_to_new_mgf = self.output_basename + "_unified.mgf"
        parse_mgf(
            peprec.df,
            self.passed_mgf_path,
            outname=path_to_new_mgf,
            filename_col="Raw file",
            spec_title_col="spec_id",
            title_parsing_method="run.scan.scan",

        )
        self._path_to_new_mgf = path_to_new_mgf

    def get_peprec(self, parse_mgf: bool = True) -> PeptideRecord:
        """Get PeptideRecord from msms.txt file, optionally parse MGF files into one."""
        logger.debug("Converting MaxQuant msms.txt to PEPREC...")
        peprec = self.msms_df.msms.to_peprec(
            modification_mapping=self._modification_mapping,
            fixed_modifications=self._fixed_modifications,
        )
        if parse_mgf:
            self.parse_mgf_files(peprec)
        return peprec

    def get_search_engine_features(self) -> pd.DataFrame:
        """Get pandas.DataFrame with search engine features."""
        return self.msms_df.msms.get_search_engine_features()


class PeptideShakerPipeline(_Pipeline):
    """PeptideShaker extended PSM report to PeptideRecord and search engine features."""

    def __init__(self, config: Dict, output_basename: Union[str, os.PathLike]) -> None:
        super().__init__(config, output_basename)

        # Private attributes, specific to this pipeline
        self._extended_psm_report = None

    @property
    def original_pin(self):
        """Get PercolatorIn object from identification file."""
        raise NotImplementedError(
            "Property `original_pin` is not implemented in class "
            "`PeptideShakerPipeline`."
        )

    def peprec_from_pin(self):
        """Get PeptideRecord from PIN file and MGF file."""
        raise NotImplementedError(
            "Method `peprec_from_pin` is not implemented in class "
            "`PeptideShakerPipeline`."
        )

    @property
    def extended_psm_report(self):
        """Get Extended PSM Report with identification results."""
        if self._extended_psm_report is None:
            self._extended_psm_report = pd.DataFrame.ext_psm_report.from_file(
                self.path_to_id_file
            )
        return self._extended_psm_report

    def get_peprec(self) -> PeptideRecord:
        """Get PeptideRecord."""
        return self.extended_psm_report.ext_psm_report.to_peprec()

    def get_search_engine_features(self) -> pd.DataFrame:
        """Get pandas.DataFrame with search engine features."""
        return self.extended_psm_report.ext_psm_report.get_search_engine_features()


class PeaksPipeline(_Pipeline):
    # TODO: move code to separate file and create mzid class
    """Peaks mzid report to PeptideRecord and search engine features."""

    def __init__(self, config: Dict, output_basename: Union[str, os.PathLike]) -> None:
        super().__init__(config, output_basename)

        # Private attributes, specific to this pipeline
        self.df = None

    @property
    def original_pin(self):
        """Get PercolatorIn object from identification file."""
        raise NotImplementedError(
            "Property `original_pin` is not implemented in class `PeaksPipeline`."
        )

    def peprec_from_pin(self):
        """Get PeptideRecord from PIN file and MGF file."""
        raise NotImplementedError(
            "Method `peprec_from_pin` is not implemented in class " "`PeaksPipeline`."
        )

    @staticmethod
    def _get_peprec_modifications(modifications: List):
        # TODO: Add unit tests for this function
        """get peprec modifications out of the peaks id file."""
        suffix_list = ["Phospho", "Dioxidation"]
        if isinstance(modifications, List):
            mods = []
            for m in modifications:
                if m["name"] in suffix_list:
                    mods.append(f"{m['location']}|{m['name'] + m['residues'][0]}")
                elif "residues" not in m.keys() and m["location"] != 0:
                    mods.append(f"-1|{m['name']}")
                else:
                    mods.append(f"{m['location']}|{m['name']}")
            return "|".join(mods)
        else:
            raise TypeError("`modifications` should be of type list.")

    def _convert_to_flat_dict(self, nested_dict, parent_key="", sep="_"):
        """Convert nested dict to flat dict with concatenated keys."""

        items = []
        for k, v in nested_dict.items():
            new_key = parent_key + sep + k if parent_key else k

            if isinstance(v, Dict):
                items.extend(self._convert_to_flat_dict(v, new_key, sep=sep))
            elif isinstance(v, List):
                if isinstance(v[0], Dict) and k != "Modification":
                    items.extend(self._convert_to_flat_dict(v[0], new_key, sep=sep))
                elif isinstance(v[0], Dict) and k == "Modification":
                    items.append((new_key, v))
                else:
                    items.append((new_key, v[0]))
            else:
                items.append((new_key, v))
        return items

    def read_df_from_mzid(self) -> pd.DataFrame:
        """Read mzid to Dataframe."""
        logger.info("Processing mzid file")
        df = pd.DataFrame(
            columns=[
                "spec_id",
                "peptide",
                "peptide_length",
                "modifications",
                "charge",
                "PEAKS:peptideScore",
                "Label",
                "Raw file",
            ]
        )
        with mzid.read(self.path_to_id_file) as reader:
            for spectrum_identification_result in tqdm(reader):
                retrieved_data = pd.Series(
                    index=[
                        "spec_id",
                        "peptide",
                        "peptide_length",
                        "modifications",
                        "charge",
                        "protein_list",
                        "PEAKS:peptideScore",
                        "Label",
                        "Raw file",
                    ]
                )
                flat_dict = dict(
                    self._convert_to_flat_dict(spectrum_identification_result)
                )
                spec_id = (
                    flat_dict["location"]
                    .rsplit("/", 1)[1]
                    .split(".", 1)[0]
                    .replace(",", "_", 1)
                    + ":"
                    + flat_dict["spectrumID"]
                )
                retrieved_data["spec_id"] = spec_id
                retrieved_data["peptide"] = flat_dict[
                    "SpectrumIdentificationItem_PeptideSequence"
                ]
                retrieved_data["peptide_length"] = len(retrieved_data["peptide"])
                try:
                    retrieved_data["modifications"] = self._get_peprec_modifications(
                        flat_dict["SpectrumIdentificationItem_Modification"]
                    )
                except KeyError:
                    retrieved_data["modifications"] = "-"
                retrieved_data["charge"] = flat_dict[
                    "SpectrumIdentificationItem_chargeState"
                ]
                retrieved_data["protein_list"] = [
                    d["accession"]
                    for d in spectrum_identification_result[
                        "SpectrumIdentificationItem"
                    ][0]["PeptideEvidenceRef"]
                    if "accession" in d.keys()
                ]
                retrieved_data["PEAKS:peptideScore"] = flat_dict[
                    "SpectrumIdentificationItem_PEAKS:peptideScore"
                ]
                retrieved_data["Label"] = flat_dict[
                    "SpectrumIdentificationItem_PeptideEvidenceRef_isDecoy"
                ]
                retrieved_data["Raw file"] = (
                    flat_dict["location"]
                    .rsplit("/", 1)[1]
                    .split(".", 1)[0]
                    .replace(",", "_", 1)
                )
                df = df.append(retrieved_data, ignore_index=True)
            df["Label"] = df["Label"].apply(lambda x: -1 if x else 1)
            return df

    def parse_mgf_files(self, peprec):
        """Parse multiple MGF files into one for MS²PIP."""
        logger.debug("Parsing MGF files into one for MS²PIP")
        path_to_new_mgf = self.output_basename + "_unified.mgf"
        parse_mgf(
            peprec,
            self.passed_mgf_path,
            outname=path_to_new_mgf,
            filename_col="Raw file",
            spec_title_col="spec_id",
            title_parsing_method="full_run",
        )
        self.passed_mgf_path = path_to_new_mgf

    def get_peprec(self) -> PeptideRecord:
        """Get PeptideRecord."""

        if not self.df:
            self.df = self.read_df_from_mzid()
        peprec_df = self.df[
            [
                "spec_id",
                "peptide",
                "modifications",
                "charge",
                "protein_list",
                "PEAKS:peptideScore",
                "Label",
                "Raw file",
            ]
        ].rename({"PEAKS:peptideScore": "psm_score"}, axis=1)
        self.parse_mgf_files(peprec_df)
        titles, rt = parse_mgf_title_rt(self.passed_mgf_path)
        id_rt_dict = {
            "spec_id": list(titles.values()),
            "observed_retention_time": list(rt.values()),
        }
        id_rt_df = pd.DataFrame.from_dict(id_rt_dict)
        peprec_df = pd.merge(peprec_df, id_rt_df, on="spec_id", how="inner")
        return PeptideRecord.from_dataframe(peprec_df)

    def get_search_engine_features(self) -> pd.DataFrame:
        """Get pandas.DataFrame with search engine features."""

        peprec_cols = [
            "peptide",
            "modifications",
            "observed_retention_time",
            "Raw file",
            "protein_list",
            "Label",
        ]

        feature_cols = [col for col in self.df.columns if col not in peprec_cols]
        print(feature_cols)
        return self.df[feature_cols]



"""Parse configuration from command line arguments and configuration files."""

import argparse
import json
import multiprocessing as mp
import os
from typing import Dict, Optional

from cascade_config import CascadeConfig

from ms2rescore import package_data
from ms2rescore._exceptions import MS2RescoreConfigurationError
from ms2rescore._version import __version__

try:
    import importlib.resources as pkg_resources
except ImportError:
    import importlib_resources as pkg_resources


def _parse_arguments() -> argparse.Namespace:
    """Parse CLI arguments."""
    parser = argparse.ArgumentParser(
        description="MS²ReScore: Sensitive PSM rescoring with predicted MS²\
            peak intensities."
    )
    parser.add_argument("-v", "--version", action="version", version=__version__)
    parser.add_argument(
        "identification_file",
        type=str,
        help="path to identification file (pin, mzid, msms.txt, tandem xml...)",
    )
    parser.add_argument(
        "-m",
        metavar="FILE",
        action="store",
        type=str,
        dest="mgf_path",
        help="path to MGF file or directory with MGF files (default: derived from\
            identification file)",
    )
    parser.add_argument(
        "-c",
        metavar="FILE",
        action="store",
        type=str,
        dest="config_file",
        help="path to MS²ReScore configuration file (see README.md)",
    )
    parser.add_argument(
        "-t",
        metavar="PATH",
        action="store",
        type=str,
        dest="tmp_path",
        help="path to directory to place temporary files"
    )
    parser.add_argument(
        "-o",
        metavar="FILE",
        action="store",
        type=str,
        dest="output_filename",
        help="name for output files (default: derive from identification file)",
    )
    parser.add_argument(
        "-l",
        metavar="LEVEL",
        action="store",
        type=str,
        dest="log_level",
        default="info",
        help="logging level (default: `info`)",
    )
    parser.add_argument(
        "-n",
        metavar="VALUE",
        action="store",
        type=int,
        dest="num_cpu",
        default=None,
        help="number of cpus available to MS²Rescore",
    )
    parser.add_argument(
        "--pipeline",
        metavar="PIPELINE",
        action="store",
        type=str,
        dest="pipeline",
        default=None,
        help="determines which pipeline to use (default: 'infer')",
    )
    parser.add_argument(
        "--percolator",
        metavar="BOOL",
        action="store",
        type=bool,
        dest="run_percolator",
        default=None,
        help="run percolator (default: 'True')",
    )
    parser.add_argument(
        "--plotting",
        metavar="BOOL",
        action="store",
        type=bool,
        dest="plotting",
        default=None,
        help="write PDF with result plots",
    )
    return parser.parse_args()


def _validate_filenames(config: Dict) -> Dict:
    """Validate and infer input/output filenames."""
    # identification_file should exist
    id_file = config["general"]["identification_file"]
    if not os.path.isfile(id_file):
        raise FileNotFoundError(id_file)

    # MGF path should either be None, or existing path to file or dir
    mgf_path = config["general"]["mgf_path"]
    if mgf_path:
        if not os.path.exists(mgf_path):
            raise FileNotFoundError(mgf_path)

    # Output filename should be None or its path should exist. If not, make path.
    if config["general"]["output_filename"]:
        output_path = os.path.abspath(config["general"]["output_filename"])
        if not os.path.isdir(output_path):
            os.makedirs(output_path, exist_ok=True)
    else:
        config["general"]["output_filename"] = os.path.splitext(id_file)[0]

    return config


def _validate_num_cpu(config: Dict) -> Dict:
    """Validate requested num_cpu with available cpu count."""
    n_available = mp.cpu_count()
    if (config["general"]["num_cpu"] == -1) or (
        config["general"]["num_cpu"] > n_available
    ):
        config["general"]["num_cpu"] = n_available
    return config


def parse_config(parse_cli_args: bool = True, config_class: Optional[Dict] = None) -> Dict:
    """
    Parse and validate MS²ReScore configuration files and arguments.

    Default configuration, user configuration files, and CLI/class arguments are parsed
    in cascading order.

    Parameters
    ----------
    parse_cli_args : bool
        parse command line arguments or not, default True
    config_class : Dict
        dictionary with arguments from the Python class; required if `parse_cli_args`
        is False
    """
    config_schema = pkg_resources.open_text(package_data, "config_schema.json")
    config_default = pkg_resources.open_text(package_data, "config_default.json")

    # MS²ReScore can be run from the CLI, or as a Python module
    if parse_cli_args:
        args = _parse_arguments()
        config_user = args.config_file
        if config_class:
            raise MS2RescoreConfigurationError(
                "If `parse_cli_args` is True, `config_class` must be None."
            )
    elif config_class:
        args = None
        config_user = config_class["general"]["config_file"]
    else:
        raise MS2RescoreConfigurationError(
            "If `parse_cli_args` is False, `config_class` arguments are required."
        )

    cascade_conf = CascadeConfig(validation_schema=json.load(config_schema))
    cascade_conf.add_dict(json.load(config_default))
    if config_user:
        cascade_conf.add_json(config_user)
    if parse_cli_args:
        cascade_conf.add_namespace(args, subkey="general")
    elif config_class:
        cascade_conf.add_dict(config_class)
    config = cascade_conf.parse()

    config = _validate_filenames(config)
    config = _validate_num_cpu(config)

    config["general"]["pipeline"] = config["general"]["pipeline"].lower()

    return config



"""Parse MGF files."""

import logging
import mmap
import os.path
import re
import random
from tqdm import tqdm

from ms2rescore._exceptions import MS2RescoreError

logger = logging.getLogger(__name__)


class ParseMGFError(MS2RescoreError):
    """Error parsing MGF file."""

    pass


def get_num_lines(file_path):
    fp = open(file_path, "r+")
    buf = mmap.mmap(fp.fileno(), 0)
    lines = 0
    while buf.readline():
        lines += 1
    return lines


def title_parser(line, method='full', run=None):
    """
    Take an MGF TITLE line and return the spectrum title.

    Depending on the software tool that was used to write the MGF files and the
    search engine, we need to extract different parts of the TITLE field. E.g,
    for MaxQuant, everything up until the first space is required

    line: string, line from MGF file containing 'TITLE='
    method: string, one of the following:
    - 'full': take everything after 'TITLE='
    - 'full_run': take everything after 'TITLE=' and add run to title
    - 'first_space': take everything between 'TITLE=' and first space.
    - 'first_space_no_charge': take everything between 'TITLE=' and first space,
      but leave out everything after last dot. (required for MaxQuant pipeline).
    - 'run.scan.scan': Extract scan number and merge with run name (for MaxQuant IDs).
    """

    if method == 'full':
        title = line[6:].strip()
    elif method == 'full_run':
        if not run:
            raise TypeError("If `method` is `full_run`, `run` cannot be None.")
        title = run + ':' + line[6:].strip()
    elif method == 'first_space':
        title = line[6:].split(' ')[0].strip()
    elif method == 'first_space_no_charge':
        title = '.'.join(line[6:].split(' ')[0].split('.')[:-1]).strip()
    elif method == 'run.scan.scan':
        if not run:
            raise TypeError("If `method` is `run.scan.scan`, `run` cannot be None.")
        scan_m = re.match(r"TITLE=.*scan=([0-9]+).*$", line)
        if scan_m:
            scan = scan_m.group(1)
        else:
            raise ParseMGFError(
                f"Could not extract scan number from TITLE field: `{line.strip()}`"
            )
        title = '.'.join([run, scan, scan])
    else:
        raise ValueError("method '{}' is not a valid title parsing method".format(
            method
        ))
    return title


def parse_mgf(df_in, mgf_folder, outname='scan_mgf_result.mgf',
              filename_col='mgf_filename', spec_title_col='spec_id',
              title_parsing_method='full',
              show_progress_bar=True):

    if not os.path.isdir(mgf_folder):
        raise NotADirectoryError(mgf_folder)

    if df_in[spec_title_col].duplicated().any():
        logger.warning("Duplicate spec_id's found in PeptideRecord.")

    if df_in[filename_col].iloc[0][-4:].lower() == '.mgf':
        file_suffix = ''
    else:
        file_suffix = '.mgf'

    runs = df_in[filename_col].unique()
    logger.info("Parsing %i MGF files to single MGF containing all PSMs.", len(runs))

    with open(outname, 'w') as out:
        count = 0
        iterator = tqdm(runs) if show_progress_bar else runs
        for run in iterator:
            current_mgf_file = os.path.join(mgf_folder, run + file_suffix)
            spec_set = set(df_in[(df_in[filename_col] == run)][spec_title_col].values)
            # Temporary fix: replace charges in MGF with ID'ed charges
            # Until MS2PIP uses ID'ed charge instead of MGF charge
            id_charges = df_in[(df_in[filename_col] == run)].set_index('spec_id')['charge'].to_dict()

            found = False
            with open(current_mgf_file, 'r') as f:
                for line in f:
                    if 'TITLE=' in line:
                        title = title_parser(line, method=title_parsing_method, run=run)
                        if title in spec_set:
                            found = True
                            line = "TITLE=" + title + "\n"
                            out.write("BEGIN IONS\n")
                            out.write(line)
                            count += 1
                            continue
                    if 'END IONS' in line:
                        if found:
                            out.write(line + '\n')
                            found = False
                            continue
                    # Temporary fix (see above, write charge if pepmass is written)
                    if 'PEPMASS=' in line:
                        if found:
                            charge = id_charges[title]
                            out.write(line)
                            out.write("CHARGE=" + str(charge) + "+\n")
                            continue
                    if 'CHARGE=' in line:
                        continue
                    # Only print lines when spectrum is found and intensity != 0
                    if found and line[-5:] != ' 0.0\n':
                        out.write(line)

    num_expected = len(df_in[spec_title_col].unique())
    logger.debug(
        "%i/%i spectra found and written to new MGF file.", count, num_expected
    )
    if not count == num_expected:
        raise ParseMGFError("Not all PSMs could be found in the provided MGF files.")



import logging


def setup_logging(passed_level):
    log_mapping = {
        'critical': logging.CRITICAL,
        'error': logging.ERROR,
        'warning': logging.WARNING,
        'info': logging.INFO,
        'debug': logging.DEBUG,
    }

    if passed_level not in log_mapping:
        print(
            "Invalid log level. Should be one of the following: ",
            ', '.join(log_mapping.keys())
        )
        exit(1)

    logging.basicConfig(
        format='%(asctime)s // %(levelname)s // %(name)s // %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S',
        level=log_mapping[passed_level]
    )



#! python
"""MS²ReScore: Sensitive PSM rescoring with predicted MS² peak intensities and RTs."""

import logging
import os
import subprocess
import tempfile
from multiprocessing import cpu_count
from typing import Dict, Optional, Union

from pandas.errors import EmptyDataError

from ms2pip.ms2pipC import MS2PIP

from ms2rescore import id_file_parser, plotting, rescore_core, setup_logging
from ms2rescore._exceptions import MS2RescoreError
from ms2rescore._version import __version__
from ms2rescore.config_parser import parse_config
from ms2rescore.retention_time import RetentionTimeIntegration

logger = logging.getLogger(__name__)


class MS2ReScore:
    """
    MS²ReScore: Sensitive PSM rescoring with predicted MS² peak intensities and RTs.

    Parameters
    ----------
    parse_cli_args : bool, optional
        parse command line arguments, default True
    configuration : dict, optional
        dict containing general ms2rescore configuration; should at least contain
        `identification_file`; required if `parse_cli_args` is False
    set_logger : bool, optional
        set custom logger or not, default False
    """

    def __init__(
        self,
        parse_cli_args: bool = True,
        configuration: Optional[Dict] = None,
        set_logger: bool = False,
    ) -> None:
        """Initialize MS2ReScore object."""
        self.config = parse_config(
            parse_cli_args=parse_cli_args, config_class=configuration
        )

        if set_logger:
            setup_logging.setup_logging(self.config["general"]["log_level"])

        if self.config["general"]["run_percolator"]:
            self._validate_cli_dependency("percolator -h")

        logger.debug(
            "Using %i of %i available CPUs.",
            self.config["general"]["num_cpu"],
            cpu_count(),
        )

        if not self.config["general"]["tmp_path"]:
            self.tmp_path = tempfile.mkdtemp()
            self.config["general"]["tmp_path"] = self.tmp_path
        else:
            self.tmp_path = self.config["general"]["tmp_path"]
            os.makedirs(self.tmp_path, exist_ok=True)

        self.tmpfile_basepath = os.path.join(
            self.tmp_path,
            os.path.basename(
                os.path.splitext(self.config["general"]["identification_file"])[0]
            ),
        )

        selected_pipeline = self._select_pipeline()
        self.pipeline = selected_pipeline(self.config, self.tmpfile_basepath)
        logger.info("Using %s.", selected_pipeline.__name__)

    @staticmethod
    def _validate_cli_dependency(command):
        """Validate that command returns zero exit status."""
        if subprocess.getstatusoutput(command)[0] != 0:
            logger.critical(
                "`%s` returned non-zero exit status. Please verify installation.",
                command,
            )
            exit(1)

    @staticmethod
    def _infer_pipeline(identification_file: str):
        """Infer pipeline from identification file."""
        logger.debug("Inferring pipeline from identification filename...")
        if identification_file.lower().endswith(".pin"):
            pipeline = id_file_parser.PinPipeline
        elif identification_file.lower().endswith(".t.xml"):
            pipeline = id_file_parser.TandemPipeline
        elif identification_file.endswith("msms.txt"):
            pipeline = id_file_parser.MaxQuantPipeline
        elif identification_file.lower().endswith(".mzid"):
            pipeline = id_file_parser.MSGFPipeline
        else:
            raise MS2RescoreError(
                "Could not infer pipeline from identification filename. Please specify "
                "`general` > `pipeline` in your configuration file."
            )
        return pipeline

    def _select_pipeline(self):
        """Select specific rescoring pipeline."""
        if self.config["general"]["pipeline"] == "infer":
            pipeline = self._infer_pipeline(
                self.config["general"]["identification_file"]
            )
        elif self.config["general"]["pipeline"] == "pin":
            pipeline = id_file_parser.PinPipeline
        elif self.config["general"]["pipeline"] == "maxquant":
            pipeline = id_file_parser.MaxQuantPipeline
        elif self.config["general"]["pipeline"] == "msgfplus":
            pipeline = id_file_parser.MSGFPipeline
        elif self.config["general"]["pipeline"] == "tandem":
            pipeline = id_file_parser.TandemPipeline
        elif self.config["general"]["pipeline"] == "peptideshaker":
            pipeline = id_file_parser.PeptideShakerPipeline
        elif self.config["general"]["pipeline"] == "peaks":
            pipeline = id_file_parser.PeaksPipeline
        else:
            raise NotImplementedError(self.config["general"]["pipeline"])
        return pipeline

    @staticmethod
    def get_ms2pip_features(
        ms2pip_config: Dict,
        peprec_filename: Union[str, os.PathLike],
        mgf_filename: Union[str, os.PathLike],
        output_filename: Union[str, os.PathLike],
        num_cpu: int,
    ):
        """Get predicted MS² peak intensities from MS2PIP."""
        logger.info("Adding MS2 peak intensity features with MS²PIP.")
        ms2pip_config = rescore_core.make_ms2pip_config_dict(ms2pip_config)

        # Check if input files exist
        for f in [peprec_filename, mgf_filename]:
            if not os.path.isfile(f):
                raise FileNotFoundError(f)

        logger.debug("Running MS²PIP...")
        ms2pip = MS2PIP(
            peprec_filename,
            spec_file=mgf_filename,
            params=ms2pip_config,
            num_cpu=num_cpu,
            add_retention_time=False,
            compute_correlations=False,
        )
        try:
            ms2pip.run()
        finally:
            ms2pip.cleanup()
        logger.debug("MS²PIP finished.")

        logger.info("Calculating features from predicted spectra")
        preds_filename = (
            peprec_filename.replace(".peprec", "")
            + "_"
            + ms2pip_config["ms2pip"]["model"]
            + "_pred_and_emp.csv"
        )
        rescore_core.calculate_features(
            preds_filename,
            output_filename + "_ms2pipfeatures.csv",
            num_cpu,
            show_progress_bar=False,
        )

    @staticmethod
    def get_rt_features(
        peprec_filename: Union[str, os.PathLike],
        output_filename: Union[str, os.PathLike],
        num_cpu: int,
    ):
        """Get retention time features with DeepLC."""
        logger.info("Adding retention time features with DeepLC.")
        rt_int = RetentionTimeIntegration(
            peprec_filename,
            output_filename + "_rtfeatures.csv",
            num_cpu=num_cpu,
        )
        rt_int.run()

    def _run_percolator(self):
        """Run Percolator with different feature subsets."""
        for subset in self.config["general"]["feature_sets"]:
            subname = (
                self.config["general"]["output_filename"]
                + "_"
                + "_".join(subset)
                + "_features"
            )
            percolator_cmd = "percolator "
            for op in self.config["percolator"].keys():
                percolator_cmd = percolator_cmd + "--{} {} ".format(
                    op, self.config["percolator"][op]
                )
            percolator_cmd = (
                percolator_cmd
                + "{} -m {} -M {} -w {} -v 0 -U --post-processing-tdc\n".format(
                    subname + ".pin",
                    subname + ".pout",
                    subname + ".pout_dec",
                    subname + ".weights",
                )
            )

            logger.info("Running Percolator: %s", percolator_cmd)
            subprocess.run(percolator_cmd, shell=True)
            if not os.path.isfile(subname + ".pout"):
                logger.error("Error running Percolator")

    def run(self):
        """Run MS²ReScore."""
        peprec = self.pipeline.get_peprec()
        peprec.validate_decoy_presence()
        peprec_filename = self.tmpfile_basepath + ".peprec"
        peprec.to_csv(peprec_filename)

        search_engine_features = self.pipeline.get_search_engine_features()
        search_engine_features_filename = (
            self.tmpfile_basepath + "_search_engine_features.csv"
        )
        search_engine_features.to_csv(search_engine_features_filename, index=False)

        if any("ms2pip" in fst for fst in self.config["general"]["feature_sets"]):
            self.get_ms2pip_features(
                self.config["ms2pip"],
                peprec_filename,
                self.pipeline.path_to_mgf_file,
                self.tmpfile_basepath,
                self.config["general"]["num_cpu"],
            )

        if any("rt" in fst for fst in self.config["general"]["feature_sets"]):
            self.get_rt_features(
                peprec_filename,
                self.tmpfile_basepath,
                self.config["general"]["num_cpu"],
            )

        logger.info("Generating PIN files")
        rescore_core.write_pin_files(
            peprec_filename,
            self.config["general"]["output_filename"],
            searchengine_features_path=search_engine_features_filename,
            ms2pip_features_path=self.tmpfile_basepath + "_ms2pipfeatures.csv",
            rt_features_path=self.tmpfile_basepath + "_rtfeatures.csv",
            feature_sets=self.config["general"]["feature_sets"],
        )

        if self.config["general"]["run_percolator"]:
            self._run_percolator()

            # Only use plotting module when run_percolator is true
            if self.config["general"]["plotting"]:
                logger.info("Generating Rescore plots")

                plotting.PIN(
                    peprec_filename, self.config["general"]["output_filename"]
                )

                for fset in self.config["general"]["feature_sets"]:
                    pout_file = (
                        self.config["general"]["output_filename"]
                        + "_"
                        + "_".join(fset)
                        + "_features.pout"
                    )
                    pout_decoy_file = (
                        self.config["general"]["output_filename"]
                        + "_"
                        + "_".join(fset)
                        + "_features.pout_dec"
                    )
                    try:
                        plotting.POUT(
                            pout_file,
                            pout_decoy_file,
                            self.config["general"]["output_filename"],
                            " ".join(fset)
                        )
                    except EmptyDataError:
                        logger.warn(f"Feature set: {'_'.join(fset)} returned empty pout file")
                        continue

                plotting.RescoreRecord.save_plots_to_pdf(
                    self.config["general"]["output_filename"] + "_plots.pdf",
                    FDR_thresholds=[0.01, 0.001],
                )
        if not self.config["general"]["run_percolator"] and self.config["general"]["plotting"]:
            logger.warn("To plot rescore results run_percolator should be enabled")
        logger.info("MS²ReScore finished!")



"""
Functions necessary to run the rescore algorithm. Currently supports MSGF+ with
concatenated searches.
"""

import logging
import multiprocessing
import os
import warnings
from typing import Dict, Optional, Union

import numpy as np
import pandas as pd
from scipy.stats import pearsonr, spearmanr
from sklearn.metrics import mean_squared_error as mse
from tqdm import tqdm
import itertools

from ms2rescore._exceptions import MS2RescoreError

logger = logging.getLogger(__name__)


def make_ms2pip_config_dict(options: Dict):
    """Write ms2pip configuration file based on ms2pip config dict."""

    # Parse model
    if "frag" in options:
        options["model"] = options["frag"]
    elif not "model" in options:
        options["model"] = "HCD"  # Assume HCD

    # Parse MS2 error tolerance
    if not "frag_error" in options:
        if options["model"] in ["CID", "CIDch2"]:
            options["frag_error"] = 0.8
        else:
            options["frag_error"] = 0.02  # Assume HCD / QTOF

    # Add config items
    ms2pip_config = {
        "ms2pip": {
            "model": options["model"],
            "frag_error": options["frag_error"],
            "ptm": [],
            "sptm": [],
            "gptm": [],
        }
    }

    # Parse modifications
    for mod in options["modifications"]:
        if mod["amino_acid"] is None and mod["n_term"]:
            aa = "N-term"
        elif mod["amino_acid"] is None and mod["c_term"]:
            aa = "C-term"
        else:
            aa = mod["amino_acid"]
        ms2pip_config["ms2pip"]["ptm"].append(
            ",".join([mod["name"], str(mod["mass_shift"]), "opt", aa])
        )

    return ms2pip_config

def df_to_dict(df):
    """Create easy to access dict from pred_and_emp."""
    preds_dict = {}
    preds_list = df[
        ["spec_id", "charge", "ion", "target", "prediction"]
    ].values.tolist()

    for row in preds_list:
        spec_id = row[0]
        if spec_id in preds_dict.keys():
            if row[2] in preds_dict[spec_id]["target"]:
                preds_dict[spec_id]["target"][row[2]].append(row[3])
                preds_dict[spec_id]["prediction"][row[2]].append(row[4])
            else:
                preds_dict[spec_id]["target"][row[2]] = [row[3]]
                preds_dict[spec_id]["prediction"][row[2]] = [row[4]]
        else:
            preds_dict[spec_id] = {
                "charge": row[1],
                "target": {row[2]: [row[3]]},
                "prediction": {row[2]: [row[4]]},
            }
    return preds_dict


def compute_features(df):
    """Compute ReScore features."""

    preds_dict = df_to_dict(df)

    rescore_features = []
    spec_ids = []
    charges = []

    feature_names = [
        "spec_pearson_norm",
        "ionb_pearson_norm",
        "iony_pearson_norm",
        "spec_mse_norm",
        "ionb_mse_norm",
        "iony_mse_norm",
        "min_abs_diff_norm",
        "max_abs_diff_norm",
        "abs_diff_Q1_norm",
        "abs_diff_Q2_norm",
        "abs_diff_Q3_norm",
        "mean_abs_diff_norm",
        "std_abs_diff_norm",
        "ionb_min_abs_diff_norm",
        "ionb_max_abs_diff_norm",
        "ionb_abs_diff_Q1_norm",
        "ionb_abs_diff_Q2_norm",
        "ionb_abs_diff_Q3_norm",
        "ionb_mean_abs_diff_norm",
        "ionb_std_abs_diff_norm",
        "iony_min_abs_diff_norm",
        "iony_max_abs_diff_norm",
        "iony_abs_diff_Q1_norm",
        "iony_abs_diff_Q2_norm",
        "iony_abs_diff_Q3_norm",
        "iony_mean_abs_diff_norm",
        "iony_std_abs_diff_norm",
        "dotprod_norm",
        "dotprod_ionb_norm",
        "dotprod_iony_norm",
        "cos_norm",
        "cos_ionb_norm",
        "cos_iony_norm",
        "spec_pearson",
        "ionb_pearson",
        "iony_pearson",
        "spec_spearman",
        "ionb_spearman",
        "iony_spearman",
        "spec_mse",
        "ionb_mse",
        "iony_mse",
        "min_abs_diff_iontype",
        "max_abs_diff_iontype",
        "min_abs_diff",
        "max_abs_diff",
        "abs_diff_Q1",
        "abs_diff_Q2",
        "abs_diff_Q3",
        "mean_abs_diff",
        "std_abs_diff",
        "ionb_min_abs_diff",
        "ionb_max_abs_diff",
        "ionb_abs_diff_Q1",
        "ionb_abs_diff_Q2",
        "ionb_abs_diff_Q3",
        "ionb_mean_abs_diff",
        "ionb_std_abs_diff",
        "iony_min_abs_diff",
        "iony_max_abs_diff",
        "iony_abs_diff_Q1",
        "iony_abs_diff_Q2",
        "iony_abs_diff_Q3",
        "iony_mean_abs_diff",
        "iony_std_abs_diff",
        "dotprod",
        "dotprod_ionb",
        "dotprod_iony",
        "cos",
        "cos_ionb",
        "cos_iony",
    ]

    # Suppress RuntimeWarnings about invalid values
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        for spec_id, preds in preds_dict.items():
            spec_ids.append(spec_id)
            charges.append(preds["charge"])

            # Create numpy arrays
            target_b = np.array(preds["target"]["B"])
            target_y = np.array(preds["target"]["Y"])
            target_all = np.concatenate([target_b, target_y])
            prediction_b = np.array(preds["prediction"]["B"])
            prediction_y = np.array(preds["prediction"]["Y"])
            prediction_all = np.concatenate([prediction_b, prediction_y])

            target_b_unlog = 2 ** target_b - 0.001
            target_y_unlog = 2 ** target_y - 0.001
            target_all_unlog = 2 ** target_all - 0.001
            prediction_b_unlog = 2 ** prediction_b - 0.001
            prediction_y_unlog = 2 ** prediction_y - 0.001
            prediction_all_unlog = 2 ** prediction_all - 0.001

            # Calculate absolute differences
            abs_diff_b = np.abs(target_b - prediction_b)
            abs_diff_y = np.abs(target_y - prediction_y)
            abs_diff_all = np.abs(target_all - prediction_all)

            abs_diff_b_unlog = np.abs(target_b_unlog - prediction_b_unlog)
            abs_diff_y_unlog = np.abs(target_y_unlog - prediction_y_unlog)
            abs_diff_all_unlog = np.abs(target_all_unlog - prediction_all_unlog)

            # Add features
            feats = np.array(
                [
                    # spec_id,
                    # preds['charge'],
                    # Features between spectra in log space
                    pearsonr(target_all, prediction_all)[0],  # Pearson all ions
                    pearsonr(target_b, prediction_b)[0],  # Pearson b ions
                    pearsonr(target_y, prediction_y)[0],  # Pearson y ions
                    mse(target_all, prediction_all),  # MSE all ions
                    mse(target_b, prediction_b),  # MSE b ions
                    mse(target_y, prediction_y),  # MSE y ions
                    np.min(abs_diff_all),  # min_abs_diff_norm
                    np.max(abs_diff_all),  # max_abs_diff_norm
                    np.quantile(abs_diff_all, 0.25),  # abs_diff_Q1_norm
                    np.quantile(abs_diff_all, 0.5),  # abs_diff_Q2_norm
                    np.quantile(abs_diff_all, 0.75),  # abs_diff_Q3_norm
                    np.mean(abs_diff_all),  # mean_abs_diff_norm
                    np.std(abs_diff_all),  # std_abs_diff_norm
                    np.min(abs_diff_b),  # ionb_min_abs_diff_norm
                    np.max(abs_diff_b),  # ionb_max_abs_diff_norm
                    np.quantile(abs_diff_b, 0.25),  # ionb_abs_diff_Q1_norm
                    np.quantile(abs_diff_b, 0.5),  # ionb_abs_diff_Q2_norm
                    np.quantile(abs_diff_b, 0.75),  # ionb_abs_diff_Q3_norm
                    np.mean(abs_diff_b),  # ionb_mean_abs_diff_norm
                    np.std(abs_diff_b),  # ionb_std_abs_diff_norm
                    np.min(abs_diff_y),  # iony_min_abs_diff_norm
                    np.max(abs_diff_y),  # iony_max_abs_diff_norm
                    np.quantile(abs_diff_y, 0.25),  # iony_abs_diff_Q1_norm
                    np.quantile(abs_diff_y, 0.5),  # iony_abs_diff_Q2_norm
                    np.quantile(abs_diff_y, 0.75),  # iony_abs_diff_Q3_norm
                    np.mean(abs_diff_y),  # iony_mean_abs_diff_norm
                    np.std(abs_diff_y),  # iony_std_abs_diff_norm
                    np.dot(target_all, prediction_all),  # Dot product all ions
                    np.dot(target_b, prediction_b),  # Dot product b ions
                    np.dot(target_y, prediction_y),  # Dot product y ions
                    np.dot(target_all, prediction_all)
                    / (
                        np.linalg.norm(target_all, 2)
                        * np.linalg.norm(prediction_all, 2)
                    ),  # Cos similarity all ions
                    np.dot(target_b, prediction_b)
                    / (
                        np.linalg.norm(target_b, 2) * np.linalg.norm(prediction_b, 2)
                    ),  # Cos similarity b ions
                    np.dot(target_y, prediction_y)
                    / (
                        np.linalg.norm(target_y, 2) * np.linalg.norm(prediction_y, 2)
                    ),  # Cos similarity y ions
                    # Same features in normal space
                    pearsonr(target_all_unlog, prediction_all_unlog)[
                        0
                    ],  # Pearson all ions
                    pearsonr(target_b_unlog, prediction_b_unlog)[0],  # Pearson b ions
                    pearsonr(target_y_unlog, prediction_y_unlog)[0],  # Pearson y ions
                    spearmanr(target_all_unlog, prediction_all_unlog)[
                        0
                    ],  # Spearman all ions
                    spearmanr(target_b_unlog, prediction_b_unlog)[0],  # Spearman b ions
                    spearmanr(target_y_unlog, prediction_y_unlog)[0],  # Spearman y ions
                    mse(target_all_unlog, prediction_all_unlog),  # MSE all ions
                    mse(target_b_unlog, prediction_b_unlog),  # MSE b ions
                    mse(target_y_unlog, prediction_y_unlog),  # MSE y ions,
                    0
                    if np.min(abs_diff_b_unlog) <= np.min(abs_diff_y_unlog)
                    else 1,  # Ion type with min absolute difference
                    0
                    if np.max(abs_diff_b_unlog) >= np.max(abs_diff_y_unlog)
                    else 1,  # Ion type with max absolute difference
                    np.min(abs_diff_all_unlog),  # min_abs_diff
                    np.max(abs_diff_all_unlog),  # max_abs_diff
                    np.quantile(abs_diff_all_unlog, 0.25),  # abs_diff_Q1
                    np.quantile(abs_diff_all_unlog, 0.5),  # abs_diff_Q2
                    np.quantile(abs_diff_all_unlog, 0.75),  # abs_diff_Q3
                    np.mean(abs_diff_all_unlog),  # mean_abs_diff
                    np.std(abs_diff_all_unlog),  # std_abs_diff
                    np.min(abs_diff_b_unlog),  # ionb_min_abs_diff
                    np.max(abs_diff_b_unlog),  # ionb_max_abs_diff_norm
                    np.quantile(abs_diff_b_unlog, 0.25),  # ionb_abs_diff_Q1
                    np.quantile(abs_diff_b_unlog, 0.5),  # ionb_abs_diff_Q2
                    np.quantile(abs_diff_b_unlog, 0.75),  # ionb_abs_diff_Q3
                    np.mean(abs_diff_b_unlog),  # ionb_mean_abs_diff
                    np.std(abs_diff_b_unlog),  # ionb_std_abs_diff
                    np.min(abs_diff_y_unlog),  # iony_min_abs_diff
                    np.max(abs_diff_y_unlog),  # iony_max_abs_diff
                    np.quantile(abs_diff_y_unlog, 0.25),  # iony_abs_diff_Q1
                    np.quantile(abs_diff_y_unlog, 0.5),  # iony_abs_diff_Q2
                    np.quantile(abs_diff_y_unlog, 0.75),  # iony_abs_diff_Q3
                    np.mean(abs_diff_y_unlog),  # iony_mean_abs_diff
                    np.std(abs_diff_y_unlog),  # iony_std_abs_diff
                    np.dot(
                        target_all_unlog, prediction_all_unlog
                    ),  # Dot product all ions
                    np.dot(target_b_unlog, prediction_b_unlog),  # Dot product b ions
                    np.dot(target_y_unlog, prediction_y_unlog),  # Dot product y ions
                    np.dot(target_all_unlog, prediction_all_unlog)
                    / (
                        np.linalg.norm(target_all_unlog, 2)
                        * np.linalg.norm(prediction_all_unlog, 2)
                    ),  # Cos similarity all ions
                    np.dot(target_b_unlog, prediction_b_unlog)
                    / (
                        np.linalg.norm(target_b_unlog, 2)
                        * np.linalg.norm(prediction_b_unlog, 2)
                    ),  # Cos similarity b ions
                    np.dot(target_y_unlog, prediction_y_unlog)
                    / (
                        np.linalg.norm(target_y_unlog, 2)
                        * np.linalg.norm(prediction_y_unlog, 2)
                    ),  # Cos similarity y ions
                ],
                dtype=np.float64,
            )

            rescore_features.append(feats)

    rescore_features = np.vstack(rescore_features)
    rescore_features = pd.DataFrame(rescore_features, columns=feature_names)
    rescore_features["spec_id"] = spec_ids
    rescore_features["charge"] = charges

    return rescore_features


def calculate_features(
    path_to_pred_and_emp, path_to_out, num_cpu, show_progress_bar=True
):
    """
    parallelize calculation of features and write them into a csv file
    """

    logger.debug("Reading MS2PIP predictions file")
    df = pd.read_csv(path_to_pred_and_emp)
    df[["prediction", "target"]] = df[["prediction", "target"]].clip(
        lower=np.log2(0.001)
    )

    logger.debug("Computing features")
    if len(df["spec_id"].unique()) < 10000:
        all_results = compute_features(df)
    else:
        # Split up df into list of chunk_size df's (will be divided over num_cpu)
        chunk_size = 100
        spec_ids = list(df["spec_id"].unique())
        split_df = [
            df[
                df["spec_id"].isin(
                    spec_ids[
                        i
                        * len(spec_ids)
                        // chunk_size : (i + 1)
                        * len(spec_ids)
                        // chunk_size
                    ]
                )
            ]
            for i in range(chunk_size)
        ]

        # Use imap, so we can use a tqdm progress bar
        with multiprocessing.Pool(int(num_cpu)) as p:
            if show_progress_bar:
                all_results = list(
                    tqdm(p.imap(compute_features, split_df), total=chunk_size)
                )
            else:
                all_results = list(p.imap(compute_features, split_df))
        all_results = pd.concat(all_results)

    logger.debug("Writing to file")
    all_results.to_csv(path_to_out, index=False)


def redo_pin_tabs(pin_filename):
    """
    Replaces triple pipe (`|||`) with tab in given file.
    """
    with open(pin_filename, "rt") as pin_in:
        with open(pin_filename + "_tmp", "wt") as pin_out:
            for line in pin_in:
                pin_out.write(line.replace("|||", "\t"))
    os.remove(pin_filename)
    os.rename(pin_filename + "_tmp", pin_filename)


def write_pin_files(
    peprec_path: str,
    savepath: str,
    searchengine_features_path: Optional[str] = None,
    ms2pip_features_path: Optional[str] = None,
    rt_features_path: Optional[str] = None,
    feature_sets: Optional[list] = None,
):
    """
    Write Percolator IN file.

    Write PIN files for each requested feature_set: search_engine_features,
    ms2pip_features, rt_features and all_features.

    Parameters
    ----------
    peprec_path: str
        Path to PEPREC file.
    save_path: str
        Directory and basename for PIN files.
    searchengine_features_path: {str, None}
        Path to CSV with search engine features.
    ms2pip_features_path: {str, None}
        Path to CSV with ms2pip features.
    rt_features_path: {str, None}
        Path to CSV with rt features.
    feature_sets: {list, None}
        Feature sets for which to write PIN files. Can contain `search_engine`,
        `ms2pip`, and `rt`.

    """
    feature_config = {"searchengine": False, "ms2pip": False, "rt": False}

    # TODO: Fix duality in `observed_retention_time`: peprec column, also rt feature
    peprec_cols = [
        "spec_id",
        "peptide",
        "peptide_peprec",
        "modifications",
        "charge",
        "label",
        "psm_score",
        "observed_retention_time",
    ]
    pin_columns = [
        "SpecId",
        "ScanNr",
        "Label",
        "Peptide",
        "ModPeptide",
        "Proteins",
        "TITLE",
    ]
    # Define which features need to be calculated
    if not feature_sets:
        feature_config = dict.fromkeys(feature_config, True)
    else:
        for feature in set(list(itertools.chain.from_iterable(feature_sets))):
            try:
                feature_config[feature] = True
            except KeyError:
                raise MS2RescoreError(f'Feature "{feature}" not available')

    # Read search engine features & Get list with feature names split by type of feature
    if feature_config["searchengine"]:
        searchengine_features = pd.read_csv(
            searchengine_features_path, sep=",", index_col=None
        )
        search_engine_feature_names = [
            col
            for col in searchengine_features.columns
            if (col not in peprec_cols) and (col not in pin_columns)
        ]
    else:
        searchengine_features = None
        search_engine_feature_names = []

    # Read MS²PIP features
    if feature_config["ms2pip"]:
        ms2pip_features = pd.read_csv(ms2pip_features_path, sep=",", index_col=None)
        ms2pip_feature_names = [
            col
            for col in ms2pip_features.columns
            if (col not in peprec_cols) and (col not in pin_columns)
        ]
    else:
        ms2pip_features = None
        ms2pip_feature_names = []

    # Read RT features
    if feature_config["rt"]:
        rt_features = pd.read_csv(rt_features_path, sep=",", index_col=None)
        rt_feature_names = ["observed_retention_time"] + [
            col
            for col in rt_features.columns
            if (col not in peprec_cols) and (col not in pin_columns)
        ]
    else:
        rt_features = None
        rt_feature_names = []

    # Read PEPREC
    with open(peprec_path, "rt") as f:
        line = f.readline()
        if line[:7] != "spec_id":
            logger.critical("PEPREC file should start with header column")
            exit(1)
        sep = line[7]

    # Convert Protein literal list to pseudo-PIN notation:
    # not yet tab-separated, but pipe-separated
    # to avoid isues in pd.DataFrame.to_csv()
    protein_converter = lambda prot_list: "|||".join(
        prot_list.strip("[]'").split("', '")
    )
    pep = pd.read_csv(
        peprec_path,
        sep=sep,
        index_col=False,
        dtype={"spec_id": str, "modifications": str},
        converters={"Proteins": protein_converter, "protein_list": protein_converter},
    )

    pep["modifications"].fillna("-", inplace=True)
    pep = pep.replace(-np.inf, 0.0).replace(np.inf, 0.0)

    # Prepare Proteins column for PIN
    # If no Proteins column in PEPREC, fill with peptide
    if "Proteins" not in pep.columns:
        if "protein_list" in pep.columns:
            pep["Proteins"] = pep["protein_list"]
        elif "ModPeptide" in pep.columns:
            pep["Proteins"] = pep["ModPeptide"]
        else:
            pep["Proteins"] = pep["peptide"]

    # ModPeptide contains sequence with modifications (e.g. from MaxQuant)
    if "ModPeptide" not in pep.columns:
        pep["ModPeptide"] = pep["peptide"]
    pep.rename(
        columns={
            "peptide": "peptide_peprec",
            "ModPeptide": "Peptide",
            "label": "Label",
        },
        inplace=True,
    )

    # Merge features and peprec DataFrames
    complete_df = pep
    for feature_set in [searchengine_features, ms2pip_features, rt_features]:
        if isinstance(feature_set, pd.DataFrame):
            on_cols = ["spec_id", "charge"]
            cols_to_use = (
                set(feature_set.columns)
                .difference(set(complete_df.columns))
                .union(set(on_cols))
            )
            # If spec_id and charge in feature_set columns, use merge
            if all(c in feature_set.columns for c in on_cols):
                complete_df = pd.merge(
                    complete_df, feature_set[cols_to_use], on=on_cols
                )
            # Else, just use concat
            else:
                if not len(complete_df) == len(feature_set):
                    raise MS2RescoreError("Feature sets do not match.")
                complete_df = pd.concat(
                    [
                        complete_df.reset_index(drop=True),
                        feature_set.reset_index(drop=True),
                    ],
                    axis=1,
                )
    complete_df = complete_df.fillna(value=0).reset_index(drop=True)
    complete_df = complete_df.loc[:, ~complete_df.columns.duplicated()]

    # Add missing columns if necessary
    if "ScanNr" not in complete_df.columns:
        complete_df["ScanNr"] = complete_df.index
    if "SpecId" not in complete_df.columns:
        complete_df["SpecId"] = complete_df["spec_id"]
    if "Label" not in complete_df.columns:
        complete_df["Label"] = complete_df["label"]

    # Write PIN files with ordered columns
    # From Percolator documentation:
    # SpecId <tab> Label <tab> ScanNr <tab> feature1name <tab> ... <tab> featureNname
    # <tab> Peptide <tab> Proteins

    # Write pin file for each feature set
    for fset in feature_sets:

        feature_config = dict.fromkeys(feature_config, False)
        col_names = []

        for feature in fset:
            feature_config[feature] = True

        if feature_config["ms2pip"]:
            col_names = col_names + ms2pip_feature_names

        if feature_config["rt"]:
            col_names = col_names + rt_feature_names

        if feature_config["searchengine"]:
            col_names = col_names + search_engine_feature_names

        outname = "_".join(fset)
        complete_df.sort_values("SpecId", axis=0, inplace=True)
        complete_df[
            ["SpecId", "Label", "ScanNr"] + col_names + ["Peptide", "Proteins"]
        ].to_csv(
            f"{savepath}_{outname}_features.pin", sep="\t", header=True, index=False
        )
        redo_pin_tabs(f"{savepath}_{outname}_features.pin")



"""MS²ReScore: Sensitive PSM rescoring with predicted MS² peak intensities and RTs."""

import logging

from ms2rescore import MS2ReScore


logger = logging.getLogger(__name__)


def main():
    """Run MS²ReScore."""
    try:
        rescore = MS2ReScore(parse_cli_args=True, configuration=None, set_logger=True)
        rescore.run()
    except Exception:
        logger.exception("Critical error occured in MS2ReScore")


if __name__ == "__main__":
    main()



"""Add retention time related features to rescoring."""

import logging
import os
from typing import Optional, Union

import click
import pandas as pd

from ms2rescore.peptide_record import PeptideRecord

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"


logger = logging.getLogger(__name__)


class RetentionTimeIntegration:
    """Retention time integration for MS²ReScore, using DeepLC."""

    def __init__(
        self,
        peprec_path: str,
        feature_path: str,
        higher_psm_score_better: bool = True,
        calibration_set_size: Optional[Union[int, float]] = 0.20,
        num_cpu: Optional[int] = None,
    ):
        """
        Retention time integration for MS²ReScore, using DeepLC.

        Parameters
        ----------
        peprec_path: str
            Path to PEPREC file with PSMs
        feature_path: str
            Path to feature file to write features to.
        higher_psm_score_better: bool
            Wheter a higher PSM score (`psm_score` column in PEPREC) denotes a better
            score. (default: True)
        calibration_set_size: int or float
            Amount of best PSMs to use for DeepLC calibration. If this value is lower
            than the number of available PSMs, all PSMs will be used. (default: 0.20)
        num_cpu: {int, None}
            Number of processes to use in DeepLC

        Properties
        ----------
        calibration_data: pandas.DataFrame
            Get calibration peptides (N best PSMs in PEPREC).
        prediction_data: pandas.DataFrame
            Get prediction peptides.

        Methods
        -------
        run()
            Get retention time predictions for PEPREC and calculate features.

        """
        self.peprec_path = peprec_path
        self.feature_path = feature_path
        self.higher_psm_score_better = higher_psm_score_better
        self.calibration_set_size = calibration_set_size
        self.num_cpu = num_cpu

        # Until fixed upstream: https://github.com/compomics/DeepLC/issues/19
        if "NUMEXPR_MAX_THREADS" not in os.environ:
            os.environ["NUMEXPR_MAX_THREADS"] = str(self.num_cpu)

        self.peprec = None
        self.feature_df = None

        if self.peprec_path:
            self.peprec = PeptideRecord(path=self.peprec_path)

    def num_calibration_psms(self, peprec):
        """Get number of calibration PSMs given `calibration_set_size` and total number of PSMs."""
        if isinstance(self.calibration_set_size, float):
            if self.calibration_set_size > 1:
                raise ValueError("`calibration_set_size` cannot be larger than 1.")
            elif self.calibration_set_size <= 0:
                raise ValueError(
                    "`calibration_set_size` cannot be smaller than or equal to 0."
                )
            else:
                num_calibration_psms = round(
                    len(peprec) * self.calibration_set_size
                )
        elif isinstance(self.calibration_set_size, int):
            if self.calibration_set_size > len(peprec):
                logger.warning(
                    "Requested number of calibration PSMs (%s) is larger than total number "
                    "of PSMs in PEPREC (%s). Using all PSMs for calibration.",
                    self.calibration_set_size,
                    peprec,
                )
                num_calibration_psms = len(peprec)
            else:
                num_calibration_psms = self.calibration_set_size
        else:
            raise TypeError(
                "Expected float or int for `calibration_set_size`. Got "
                f"{type(self.calibration_set_size)} instead"
            )
        logger.debug("Using %i PSMs for calibration", num_calibration_psms)
        return num_calibration_psms

    def get_calibration_data(self, peprec):
        """Get calibration peptides (N best PSMs in PEPREC)."""
        ascending = not self.higher_psm_score_better
        if "label" in peprec.columns:
            label_col = "label"
        elif "Label" in peprec.columns:
            label_col = "Label"
        else:
            raise ValueError("No label column found in peptide record.")
        calibration_data = (
            peprec[peprec[label_col] == 1]
            .sort_values(["psm_score"], ascending=ascending)
            .head(self.num_calibration_psms(peprec=peprec))
            .rename(
                columns={
                    "observed_retention_time": "tr",
                    "peptide": "seq",
                }
            )[["tr", "seq", "modifications"]]
            .copy()
        )
        return calibration_data

    def get_prediction_data(self, peprec):
        """Get prediction peptides."""
        return peprec[["peptide", "modifications"]].rename(
            columns={
                "peptide": "seq",
            }
        )

    def _calculate_features(self):
        """Calculate retention time features for rescoring."""
        # Absolute difference between observed and predicted'
        self.feature_df = self.peprec.df.copy()
        self.feature_df["rt_diff"] = (
            self.feature_df["observed_retention_time"]
            - self.feature_df["predicted_retention_time"]
        ).abs()

        # Minimum RT difference for a peptidoform
        min_rt_diff = self.feature_df[
            [
                "peptide",
                "modifications",
                "observed_retention_time",
                "predicted_retention_time",
                "rt_diff",
            ]
        ].copy()

        min_rt_diff = (
            min_rt_diff.sort_values("rt_diff", ascending=True)
            .drop_duplicates(subset=["peptide", "modifications"], keep="first")
            .rename(
                columns={
                    "rt_diff": "rt_diff_best",
                    "observed_retention_time": "observed_retention_time_best",
                    "predicted_retention_time": "predicted_retention_time_best",
                }
            )
        )

        # Merging minimum RT difference features to full set
        self.feature_df = self.feature_df.merge(
            min_rt_diff, on=["peptide", "modifications"], how="left"
        )

        # Only keep feature columns
        id_columns = ["spec_id", "charge", "peptide", "modifications"]
        feature_columns = [
            "observed_retention_time",
            "predicted_retention_time",
            "rt_diff",
            "rt_diff_best",
            "observed_retention_time_best",
            "predicted_retention_time_best",
        ]
        self.feature_df = self.feature_df[id_columns + feature_columns].copy()

    def run(self):
        """Get retention time predictions for PEPREC and calculate features."""

        from deeplc import DeepLC

        if "Raw file" in self.peprec.df.columns:
            raw_specific_predicted_dfs = []
            for i, (raw_file, df) in enumerate(self.peprec.df.groupby("Raw file")):
                logger.info(f"Calibrating {raw_file}")

                peprec_raw_df = df.copy().reset_index()
                retention_time_df = pd.DataFrame(
                    columns=["spec_id", "predicted_retention_time"]
                )

                if i == 0:
                    self.deeplc_predictor = DeepLC(
                        split_cal=10,
                        n_jobs=self.num_cpu,
                        cnn_model=True,
                        verbose=False
                    )
                    self.deeplc_predictor.calibrate_preds(
                        seq_df=self.get_calibration_data(peprec_raw_df)
                    )
                    self.deeplc_model = list(self.deeplc_predictor.model.keys())
                else:
                    self.deeplc_predictor = DeepLC(
                        split_cal=10,
                        n_jobs=self.num_cpu,
                        cnn_model=True,
                        verbose=False,
                        path_model=self.deeplc_model
                    )
                    self.deeplc_predictor.calibrate_preds(
                        seq_df=self.get_calibration_data(peprec_raw_df)
                    )
                predicted_rts = pd.Series(
                    self.deeplc_predictor.make_preds(
                        seq_df=self.get_prediction_data(peprec_raw_df)
                    )
                )
                retention_time_df["spec_id"] = peprec_raw_df["spec_id"].copy()
                retention_time_df["predicted_retention_time"] = predicted_rts

                raw_specific_predicted_dfs.append(retention_time_df)
            self.peprec.df = pd.merge(
                self.peprec.df,
                pd.concat(raw_specific_predicted_dfs, ignore_index=True),
                on="spec_id",
                how="inner",
            )
        else:
            self.deeplc_predictor = DeepLC(
                split_cal=10, n_jobs=self.num_cpu, cnn_model=True, verbose=False
            )
            self.deeplc_predictor.calibrate_preds(
                seq_df=self.get_calibration_data(self.peprec.df)
            )
            predicted_rts = pd.Series(
                self.deeplc_predictor.make_preds(
                    seq_df=self.get_prediction_data(self.peprec.df)
                )
            )
            self.peprec.df["predicted_retention_time"] = predicted_rts

        self._calculate_features()
        self.feature_df.to_csv(self.feature_path, index=False)


@click.command()
@click.argument("peprec")
@click.argument("features")
def main(**kwargs):
    """Run ms2rescore.retention_time."""
    rt = RetentionTimeIntegration(kwargs["peprec"], kwargs["features"])
    rt.run()


if __name__ == "__main__":
    main()



"""MS²Rescore exceptions."""


class MS2RescoreError(Exception):
    """Generic MS2Rescore error."""

    pass


class MS2RescoreConfigurationError(MS2RescoreError):
    """Invalid MS2Rescore configuration."""

    pass



"""Interface to MaxQuant msms.txt files."""

import logging
import os
import re
from typing import Dict, List, Optional, Tuple, Union

import click
import numpy as np
import pandas as pd

from ms2rescore.peptide_record import PeptideRecord

logger = logging.getLogger(__name__)


@pd.api.extensions.register_dataframe_accessor("msms")
class MSMSAccessor:
    """Pandas extension for MaxQuant msms.txt files."""
    default_columns = {
        "Raw file",
        "Scan number",
        "Charge",
        "Length",
        "Sequence",
        "Modified sequence",
        "Proteins",
        "Missed cleavages",
        "Mass",
        "Mass error [Da]",
        "Mass error [ppm]",
        "Reverse",
        "Retention time",
        "PEP",
        "Score",
        "Delta score",
        "Localization prob",
        "Matches",
        "Intensities",
        "Mass Deviations [Da]",
        "Mass Deviations [ppm]",
        "Intensity coverage",
        "id",
    }
    _mass_error_unit = None

    def __init__(self, pandas_obj) -> None:
        """Pandas extension for MaxQuant msms.txt files."""
        self._obj = pandas_obj
        self._set_mass_error_unit()
        self.invalid_amino_acids = r"[BJOUXZ]"

    @classmethod
    def _evaluate_columns(cls, column: str) -> bool:
        """Case insensitive column evaluation for Pandas.read_csv usecols argument."""
        return column.lower() in [col.lower() for col in cls.default_columns]

    @classmethod
    def _fix_column_case(cls, columns: List[str]) -> Dict[str, str]:
        """
        Create mapping for column names with the correct case.

        Using `_evaluate_columns`, we can load required columns in a case-insensitive
        manner. As a result, the column name case must be fixed for downstream usage.
        """
        case_mapping = {col.lower(): col for col in cls.default_columns}
        rename_mapping = {col: case_mapping[col.lower()] for col in columns}
        return rename_mapping

    @classmethod
    def from_file(
        cls,
        path_to_msms: Union[str, os.PathLike],
        filter_rank1_psms: bool = True,
        validate_amino_acids: bool = True,
    ) -> pd.DataFrame:
        """
        Read msms.txt from file.

        Parameters
        ----------
        path_to_msms : str, os.Pathlike
            path to msms.txt file
        filter_rank1_psms : bool, optional
            filter for rank 1 PSMs
        validate_amino_acids : bool, optional
            remove PSMs where the sequence includes an invalid amino acid; required for
            MS2PIP compatibility

        Returns
        -------
        msms : ms2rescore.maxquant.MSMS
            MSMS object (pandas.DataFrame with additional methods)
        """

        msms_df = pd.read_csv(path_to_msms, sep="\t", usecols=cls._evaluate_columns)
        msms_df.rename(columns=cls._fix_column_case(msms_df.columns), inplace=True)
        if filter_rank1_psms:
            msms_df = msms_df.msms.filter_rank1_psms()
        if validate_amino_acids:
            msms_df = msms_df.msms.remove_invalid_amino_acids()
        return msms_df

    def _set_mass_error_unit(self) -> None:
        """Get mass error unit from DataFrame columns."""
        if "Mass error [Da]" in self._obj.columns:
            self._mass_error_unit = "Da"
        elif "Mass error [ppm]" in self._obj.columns:
            self._mass_error_unit = "ppm"
        else:
            raise NotImplementedError(f"MSMS.txt mass error unit not supported.")

    def filter_rank1_psms(self) -> pd.DataFrame:
        """Filter MSMS for rank 1 PSMs."""
        self._obj = self._obj.sort_values("Score", ascending=False)
        duplicate_indices = self._obj[
            self._obj.duplicated(["Raw file", "Scan number"], keep="first")
        ].index
        self._obj = self._obj.drop(duplicate_indices).sort_index().reset_index()

        logger.debug(
            f"Found {len(self._obj)} rank 1 PSMs of which "
            f"{len(self._obj[self._obj['Reverse'] == '+']) / len(self._obj):.0%} are "
            "decoy hits."
        )
        if len(duplicate_indices) > 0:
            logger.warning(
                "Removed %i non-rank 1 PSMs.", len(duplicate_indices)
            )

        return self._obj

    def remove_invalid_amino_acids(self) -> pd.DataFrame:
        """Remove invalid amino acids from MSMS."""
        invalid_indices = self._obj[self._obj["Sequence"].str.contains(
            self.invalid_amino_acids, regex=True
        )].index
        self._obj = self._obj.drop(index=invalid_indices).reset_index(drop=True)

        if len(invalid_indices) > 0:
            logger.warning(
                "Removed %i PSMs with invalid amino acids.", len(invalid_indices)
            )

        return self._obj

    def _get_spec_id(self) -> pd.Series:
        """Get PEPREC-style spec_id."""
        return (
            self._obj["Raw file"]
            + "."
            + self._obj["Scan number"].astype(str)
            + "."
            + self._obj["Scan number"].astype(str)
        ).rename("spec_id")

    @staticmethod
    def _get_peprec_modifications(
        modified_sequences: pd.Series,
        modification_mapping: Optional[Dict] = None,
        fixed_modifications: Optional[List] = None
    ) -> List:
        """Get peprec-formatted modifications."""
        # Process input mods
        if not modification_mapping:
            modification_mapping = dict()
        else:
            # Put braces around labels
            modification_mapping = {
                "({})".format(k): v for k, v in modification_mapping.items()
            }

        if not fixed_modifications:
            fixed_modifications = dict()
        else:
            # Match mod name to label
            modification_mapping_rev = {v: k for k, v in modification_mapping.items()}
            fixed_modifications = {
                k: modification_mapping_rev[v]
                for k, v
                in fixed_modifications.items()
            }
        # Apply fixed modifications
        if fixed_modifications:
            for aa, mod in fixed_modifications.items():
                modified_sequences = modified_sequences.str.replace(
                    aa, "{}{}".format(aa, mod)
                )

        # Parse all modifications
        parsed_modifications = [
            "|".join([
                "{}|{}".format(m.start(0) - 1 - i * 4, modification_mapping[m.group()])
                for i, m
                in enumerate(re.finditer(r"\([a-z].\)", s))
            ])
            for s in modified_sequences
        ]
        parsed_modifications = [
            "-" if mods == "" else mods for mods in parsed_modifications
        ]

        return parsed_modifications

    @staticmethod
    def _calculate_top7_peak_features(
        intensities: List,
        mass_errors: List
    ) -> Tuple[np.ndarray]:
        """
        Calculate "top 7 peak"-related search engine features.

        The following features are calculated:
        - mean_error_top7: Mean of mass errors of the seven fragment ion peaks with the
          highest intensities
        - sq_mean_error_top7: Squared MeanErrorTop7
        - stdev_error_top7: Standard deviation of mass errors of the seven fragment ion
          peaks with the highest intensities

        """
        if not (isinstance(intensities, list) and isinstance(mass_errors, list)):
            return np.nan, np.nan, np.nan

        else:
            intensities = [float(i) for i in intensities]
            mass_errors = [float(i) for i in mass_errors]

            indices_most_intens = np.array(intensities).argsort()[-1:-8:-1]
            mass_errors_top7 = [(mass_errors[i]) for i in indices_most_intens]
            mean_error_top7 = np.mean(mass_errors_top7)
            sq_mean_error_top7 = mean_error_top7 ** 2
            stdev_error_top7 = np.std(mass_errors_top7)

            return mean_error_top7, sq_mean_error_top7, stdev_error_top7

    @staticmethod
    def _calculate_ion_current_features(
        matches: List,
        intensities: List,
        intensity_coverage: List
    ) -> Tuple[np.ndarray]:
        """
        Calculate ion current related search engine features.

        The following features are calculated:
        - ln_explained_ion_current: Summed intensity of identified fragment ions,
          divided by that of all fragment ions, logged
        - ln_nterm_ion_current_ratio: Summed intensity of identified N-terminal
          fragments, divided by that of all identified fragments, logged
        - ln_cterm_ion_current_ratio: Summed intensity of identified N-terminal
          fragments, divided by that of all identified fragments, logged
        - ln_ms2_ion_current: Summed intensity of all observed fragment ions, logged

        """
        pseudo_count = 0.00001
        if not isinstance(intensities, list):
            return np.nan, np.nan, np.nan, np.nan
        else:
            ln_explained_ion_current = intensity_coverage + pseudo_count
            summed_intensities = sum([float(i) for i in intensities])

            # Calculate ratio between matched b- and y-ion intensities
            y_ion_int = sum([
                float(intensities[i])
                for i, m
                in enumerate(matches) if m.startswith("y")
            ])
            y_int_ratio = y_ion_int / summed_intensities

            ln_nterm_ion_current_ratio = (y_int_ratio + pseudo_count) * ln_explained_ion_current
            ln_cterm_ion_current_ratio = (1 - y_int_ratio + pseudo_count) * ln_explained_ion_current
            ln_ms2_ion_current = summed_intensities / ln_explained_ion_current

            out = [
                ln_explained_ion_current,
                ln_nterm_ion_current_ratio,
                ln_cterm_ion_current_ratio,
                ln_ms2_ion_current,
            ]

        return tuple([np.log(x) for x in out])

    def to_peprec(
        self,
        modification_mapping=None,
        fixed_modifications=None,
    ) -> PeptideRecord:
        """
        Get PeptideRecord from MaxQuant msms.txt file.

        Parameters
        ----------
        modification_mapping: dict
            Mapping used to convert the two-letter MaxQuant modification labels to
            PSI-MS modification names.
        fixed_modifications: dict
            Dictionary ({aa: mod}) can contain fixed modifications to be added to the
            peprec. E.g. `{'C': 'Carbamidomethyl'}`, as the MaxQuant output does not
            include modifications that were set as fixed during the search. The first
            tuple element contains the one-letter amino acid code. The second tuple
            element contains the full modification name, as listed in the values of
            `modification_mapping`.

        """
        peprec = pd.DataFrame(
            columns=[
                "spec_id",
                "peptide",
                "modifications",
                "charge",
                "protein_list",
                "psm_score",
                "observed_retention_time",
                "Label",
                "Raw file"
            ]
        )
        peprec["spec_id"] = self._get_spec_id()
        peprec["peptide"] = self._obj["Sequence"]
        peprec["modifications"] = self._get_peprec_modifications(
            self._obj["Modified sequence"],
            modification_mapping=modification_mapping,
            fixed_modifications=fixed_modifications
        )
        peprec["charge"] = self._obj["Charge"]

        # Fill NaN values in Proteins column for decoy PSMs
        # But first check that NaN Proteins only occur for decoy PSMs, if so:
        # fill these without the "REV_"
        peprec["protein_list"] = self._obj["Proteins"].str.split(";")
        if (peprec["protein_list"].isna() & self._obj["Reverse"].isna()).any():
            req_cols = zip(self._obj["Proteins"], self._obj["Reverse"], self._obj["Modified sequence"])
            peprec["protein_list"] = [
                [modseq] if (type(rev) == float) & (type(prot) == float) else prot
                for prot, rev, modseq in req_cols
            ]
        peprec["protein_list"] = peprec["protein_list"].fillna(
           "REV_" + self._obj["Modified sequence"]
        )

        peprec["psm_score"] = self._obj["Score"]
        peprec["observed_retention_time"] = self._obj["Retention time"]
        peprec["Label"] = self._obj["Reverse"].isna().apply(lambda x: 1 if x else -1)
        peprec["Raw file"] = self._obj["Raw file"]
        peprec.sort_values("spec_id", inplace=True)
        peprec.reset_index(drop=True, inplace=True)

        return PeptideRecord.from_dataframe(peprec)

    def get_search_engine_features(self):
        """
        Get search engine features from MSMS for Percolator rescoring.

        Percolator features are derived from the MSGF2PIN script. See table 1 of
        Percolator-MSGF+ article (doi.org/10.1021/pr400937n).
        """
        logger.debug("Calculating search engine features...")

        spec_id = self._get_spec_id()
        charge = self._obj["Charge"].rename("charge")

        directly_copied = self._obj[[
            "Score",
            "Delta score",
            "Localization prob",
            "Charge",
            "Mass",
            "Length",
            f"Mass error [{self._mass_error_unit}]",
            "Missed cleavages",
        ]].rename(columns={
            "Score": "RawScore",
            "Delta score": "RawDeltaScore",
            "Localization prob": "RawModLocProb",
            "Length": "PepLen",
            f"Mass error [{self._mass_error_unit}]": "dM",
            "Charge": "ChargeN",
            "Missed cleavages": "enzInt",
        })

        absdM = self._obj[f"Mass error [{self._mass_error_unit}]"].abs().rename("absdM")

        charges_encoded = pd.get_dummies(self._obj["Charge"], prefix="Charge", prefix_sep='')

        top7_features = pd.DataFrame([
            self._calculate_top7_peak_features(i, md)
            for i, md in zip(
                self._obj["Intensities"].str.split(";"),
                self._obj["Mass Deviations [Da]"].str.split(";"),
            )],
            columns=["MeanErrorTop7", "sqMeanErrorTop7", "StdevErrorTop7"],
        )

        ion_current_features = pd.DataFrame([
            self._calculate_ion_current_features(m, i, ic)
            for m, i, ic in zip(
                self._obj["Matches"].str.split(";"),
                self._obj["Intensities"].str.split(";"),
                self._obj["Intensity coverage"],
            )],
            columns=[
                "lnExplainedIonCurrent",
                "lnNTermIonCurrentRatio",
                "lnCTermIonCurrentRatio",
                "lnMS2IonCurrent",
            ],
        )

        features = pd.concat([
            spec_id,
            charge,
            directly_copied,
            absdM,
            charges_encoded,
            top7_features,
            ion_current_features,
        ], axis=1).sort_values("spec_id").reset_index(drop=True)

        return features


@click.command()
@click.argument("input-msms")
@click.argument("output-peprec")
def main(**kwargs):
    """Convert msms.txt to PEPREC."""
    msms_df = pd.DataFrame.msms.from_file(kwargs["input_psm_report"])
    peprec = msms_df.msms.to_peprec()
    peprec.to_csv(kwargs["output_peprec"])


if __name__ == "__main__":
    main()



"""Percolator integration."""


import logging
import os
import re
import subprocess
from io import StringIO
from math import floor
from typing import Dict, List, Optional, Tuple, Union

import pandas as pd

from ms2rescore._exceptions import MS2RescoreError
from ms2rescore.peptide_record import PeptideRecord


logger = logging.getLogger(__name__)


class PercolatorInError(MS2RescoreError):
    """Error while processing Percolator In file."""


class UnknownModificationLabelStyleError(PercolatorInError):
    """Could not infer modification label style."""

    pass


class UnknownModificationError(PercolatorInError):
    """Modification not found in `modification_mapping`."""

    pass


class PercolatorIn:
    """Percolator In (PIN)."""

    def __init__(
        self,
        path: Optional[str] = None,
        modification_mapping: Optional[
            Dict[Tuple[Union[str, None], Union[float, str]], str]
        ] = None,
        invalid_amino_acids: Optional[str] = r"[BJOUXZ]"
    ):
        """
        Percolator In (PIN).

        Parameters
        ----------
        path: str, optional
            Path to PIN file.
        modification_mapping: dict, optional
            Mapping of modification residue and PIN label to the modification name.
            Modification residue is the one-letter amino acid code, or None.
            Modification labels can either be strings (e.g.
            `{("Q", "UNIMOD:28"): "Gln->pyro-Glu"}`) or floats representing mass shifts
            (e.g. `{("Q", -17.02655): "Gln->pyro-Glu"}). If the keys are floats, they
            are rounded up to three decimals to avoid rounding issues while matching.
            If None, the original modification labels from the PIN file will be used.
        invalid_amino_acids: str, optional
            regex pattern of invalid amino acids. PSMs containing these amino acids will
            be dropped. (default: `r"[BJOUXZ]"`)

        """
        # Attributes
        self.modification_pattern = r"\[([^\[^\]]*)\]"
        self.invalid_amino_acids = invalid_amino_acids

        # Parameters
        self.path = path
        self.modification_mapping = modification_mapping

        if self.path:
            self.read()

    @staticmethod
    def _round(n, decimals=0):
        multiplier = 10 ** decimals
        return floor(n * multiplier + 0.5) / multiplier

    @property
    def modification_mapping(self):
        """Get modification_mapping."""
        return self._modification_mapping

    @modification_mapping.setter
    def modification_mapping(
        self, value: Optional[Dict[Tuple[Union[str, None], Union[float, str]], str]]
    ):
        """Set modification_mapping."""
        if value:
            mod_labels = [key[1] for key in value.keys()]
            if all([isinstance(label, str) for label in mod_labels]):
                self.modification_pattern = r"\[([^\[^\]]*)\]"
                self._modification_label_type = "str"
                self._modification_mapping = value
            elif all([isinstance(label, float) for label in mod_labels]):
                self.modification_pattern = r"\[([+-]?[0-9\-.]*)\]"
                self._modification_label_type = "float"
                self._modification_mapping = {
                    (aa, self._round(shift)): name
                    for (aa, shift), name in value.items()
                }
            else:
                raise TypeError(
                    "modification_mapping keys cannot have mixed types and should "
                    "either be string or float."
                )
        else:
            self.modification_pattern = r"\[([^\[^\]]*)\]"
            self._modification_mapping = None
            self._modification_label_type = None

    def modification_mapping_from_list(
        self, modification_config, label_style="infer"
    ):
        """
        Generate modification_mapping from ms2rescore configuration.

        Parameters
        ----------
        config: Dict
            ms2rescore configuration
        label_style: str
            Modification label style, any of {"infer", "unimod_accession", "mass_shift"}
            If `infer`, the style will be attempted to be infered from the PIN file.

        """
        if label_style == "infer":
            label_style = self._infer_modification_label_style()
            self.modification_mapping_from_list(
                modification_config, label_style=label_style
            )
        elif label_style == "unimod_accession":
            self.modification_mapping = {
                (mod["amino_acid"], f"UNIMOD:{mod['unimod_accession']}"): mod["name"]
                for mod in modification_config
            }
        elif label_style == "mass_shift":
            self.modification_mapping = {
                (mod["amino_acid"], mod["mass_shift"]): mod["name"]
                for mod in modification_config
            }
        else:
            raise ValueError(label_style)

    def _infer_modification_label_style(self) -> str:
        """Infer modification label style (e.g. unimod_accession or mass shift)."""
        # TODO: What happens if there are no mods? Crashes?
        modified_peptides = self.df[
            self.df["Peptide"].str.contains(r"\[([^\[^\]]*)\]", regex=True)
        ]["Peptide"]
        if (
            modified_peptides.str.extract(r"\[([^\[^\]]*)\]", expand=False)
            .str.startswith("UNIMOD:")
            .all()
        ):
            mod_notation = "unimod_accession"
        elif (
            pd.to_numeric(
                modified_peptides.str.extract(r"\[([^\[^\]]*)\]", expand=False),
                errors="coerce",
            )
            .notnull()
            .all()
        ):
            mod_notation = "mass_shift"
        else:
            raise UnknownModificationLabelStyleError()
        return mod_notation

    def _find_mods_recursively(
        self, mod_seq: str, mod_list: Optional[List[str]] = None
    ) -> List[str]:
        """
        Find modifications in modified sequence string recursively.
        """
        if not mod_list:
            mod_list = []
        match = re.search(self.modification_pattern, mod_seq)
        if match:
            mod_location = match.start()
            mod_aa = mod_seq[mod_location - 1] if mod_location > 0 else mod_seq[match.end()]
            if self.modification_mapping:
                if self._modification_label_type == "float":
                    mod_label = self._round(float(match.group(1)))
                elif self._modification_label_type == "str":
                    mod_label = match.group(1)
                else:
                    raise ValueError(
                        "Unsupported modification_label_type {}".format(
                            self._modification_label_type
                        )
                    )
                # First try with specific amino acid, then try with None (e.g. N-term)
                if (mod_aa, mod_label) in self.modification_mapping:
                    mod_name = self.modification_mapping[mod_aa, mod_label]
                elif (None, mod_label) in self.modification_mapping:
                    mod_name = self.modification_mapping[None, mod_label]
                else:
                    raise UnknownModificationError(
                        f"No modification with label `{mod_label}` for amino acid "
                        f"`{mod_aa}` found in modification_mapping "
                        f"`{self.modification_mapping}`."
                    )
                # Set C-terminal modifications location to "-1"
                if match.end() == len(mod_seq):
                    mod_location = -1
            else:
                mod_name = match.group(1)
            mod_list.extend([str(mod_location), mod_name])
            mod_seq = re.sub(self.modification_pattern, "", mod_seq, count=1)
            mod_list = self._find_mods_recursively(mod_seq, mod_list)
        return mod_list

    def _get_peprec_modifications(self, modified_sequence: str) -> str:
        """Parse modified sequence to get PEPREC-style modifications."""
        mod_seq = ".".join(modified_sequence.split(".")[1:-1])
        mod_list = self._find_mods_recursively(mod_seq)

        # In PIN, N-terminal mods are placed on position 1, instead of 0. This would
        # lead to issues in ms2pip if another modification is also present on the first
        # amino acid. Fix by setting N-term modification to 0.
        if len(mod_list) > 3:
            if mod_list[0] == "1" and mod_list[2] == "1":
                mod_list[0] = "0"

        mods = "|".join(mod_list)

        if not mods:
            mods = "-"
        return mods

    def _get_unmodified_sequence(self, modified_sequence: str) -> pd.Series:
        """Parse modified sequence to get unmodified sequence."""
        unmod_seq = ".".join(modified_sequence.split(".")[1:-1])
        unmod_seq = re.sub(self.modification_pattern, "", unmod_seq, count=0)
        return unmod_seq

    def _get_peprec_modifications_column(self) -> pd.Series:
        """Get PEPREC-style modifications column to PIN DataFrame."""
        return self.df["Peptide"].apply(self._get_peprec_modifications)

    def _get_sequence_column(self) -> pd.Series:
        """Get unmodified sequence column to PIN DataFrame."""
        return self.df["Peptide"].apply(self._get_unmodified_sequence)

    def _get_charge_column(self) -> pd.Series:
        """Get charge column from one-hot encoded `ChargeX` columns."""
        charge_cols = [col for col in self.df.columns if col.lower().startswith("charge")]
        if not (self.df[charge_cols] == 1).any(axis=1).all():
            raise PercolatorInError("Not all PSMs have an assigned charge state.")
        return (
            self.df[charge_cols]
            .rename(
                columns={col: int(col.lower().replace("charge", "")) for col in charge_cols}
            )
            .idxmax(1)
        )

    def _get_spectrum_index_column(self, pattern: Optional[str] = None) -> pd.Series:
        """Get spectrum index column, parsed from Percolator SpecId."""
        if not pattern:
            pattern = r".+_([0-9]+)_[0-9]+_[0-9]+"
        id_col = self.df["SpecId"].str.extract(pattern, expand=False).astype(int)
        if id_col.duplicated().any():
            raise PercolatorInError("Issue in matching spectrum IDs, duplicates found.")
        if id_col.isna().any():
            raise PercolatorInError("Issue in matching spectrum IDs, NaN found.")
        return id_col

    def add_peprec_modifications_column(self):
        """Add PEPREC-style modifications column to PIN DataFrame."""
        self.df["modifications"] = self._get_peprec_modifications_column()

    def add_sequence_column(self):
        """Add unmodified sequence column to PIN DataFrame."""
        self.df["sequence"] = self._get_sequence_column()

    def add_charge_column(self):
        """Add charge column from one-hot encoded `ChargeX` columns."""
        self.df["charge"] = self._get_charge_column()

    def add_spectrum_index_column(
        self, pattern: Optional[str] = None, label: Optional[str] = "spectrum_index"
    ):
        """
        Add spectrum index column, parsed from Percolator SpecId.

        See https://github.com/percolator/percolator/issues/147 for pattern explanation.
        """
        self.df[label] = self._get_spectrum_index_column(pattern=pattern)

    def get_spectrum_filename(
        self, pattern: Optional[str] = r"(.+)_[0-9]+_[0-9]+_[0-9]+"
    ) -> str:
        """
        Get spectrum filename, parsed from Percolator SpecId.

        See https://github.com/percolator/percolator/issues/147 for pattern explanation.
        """
        spectrum_filenames = (
            self.df["SpecId"].str.extract(pattern, expand=False).unique()
        )
        if len(spectrum_filenames) == 1:
            return spectrum_filenames[0]
        else:
            raise ValueError("Multiple spectrum filenames found in single PIN file.")

    def drop_invalid_amino_acids(self):
        """Drop all PSMs (rows) with peptides containing invalid amino acids."""
        if "sequence" in self.df.columns:
            sequences = self.df['sequence']
        else:
            sequences = self._get_sequence_column()
        to_drop = sequences[
            sequences.str.contains(self.invalid_amino_acids, regex=True)
        ].index
        if len(to_drop) > 0:
            logger.warning(
                "Dropping %i PSMs from PIN due to invalid amino acids (%s)",
                len(to_drop),
                self.invalid_amino_acids
            )
            self.df = self.df.drop(index=to_drop)

    @staticmethod
    def fix_tabs(
        path: str, id_column: str = "SpecId", prot_sep: Optional[str] = "|||"
    ) -> StringIO:
        """
        Return StringIO instance of PIN/POUT file with fixed Proteins column separator.

        In a PIN/POUT file, multiple proteins in the Proteins column are separated by a
        tab, which is the same separator used to separate different columns in the file.
        This makes it impossible to read with pandas. This function makes a temporary
        copy of the PIN file with the Proteins tab-separations replaced with another
        string.

        Parameters
        ----------
        path: str
            Path to input file
        id_column: str
            Label of the ID column: `SpecId` for PIN files, `PSMId` for POUT files.
        prot_sep: str
            Separator to use in proteins column.

        Returns
        -------
        io.StringIO
            io.StringIO object with fixed PIN/POUT file.

        """
        fixed_file = StringIO()
        with open(path, "rt") as file_in:
            numcol = None
            for i, line in enumerate(file_in):
                if i == 0 & line.startswith(id_column):
                    numcol = len(line.split("\t"))
                    fixed_file.write(line)
                elif i == 1 & line.startswith("DefaultDirection"):
                    pass
                    # fixed_file.write(line)
                else:
                    r = line.strip().split("\t")
                    r_cols = r[: numcol - 1]
                    r_proteins = r[numcol - 1 :]
                    r_cols.append(prot_sep.join(r_proteins))
                    fixed_file.write("\t".join(r_cols) + "\n")

        fixed_file.seek(0)
        return fixed_file

    @staticmethod
    def write_with_tabs(
        file_object: StringIO, path: str, prot_sep: Optional[str] = "|||"
    ):
        """Write PIN io.StringIO object to file with tab-separated Proteins column."""
        raise NotImplementedError

    def _get_default_direction_row(self):
        """Get default direction row from PIN file."""
        if not self.path:
            raise ValueError("PIN path is None. First set path.")
        default_direction = None
        with open(self.path, "rt") as pin_in:
            for i, line in enumerate(pin_in):
                if i == 1 & line.startswith("DefaultDirection"):
                    default_direction = line.strip().split("\t")
                if i > 1:
                    break
        return default_direction

    def read(self, path: Optional[str] = None):
        """Read PIN file into pandas.DataFrame."""
        if path:
            self.path = path
        if not self.path:
            raise ValueError("No path for PIN file defined.")
        self.df = pd.read_csv(self.fix_tabs(self.path), sep="\t")
        if self.invalid_amino_acids:
            self.drop_invalid_amino_acids()

    def write(self, path: Optional[str] = None):
        """Write PIN to file."""
        raise NotImplementedError

    def get_feature_table(self) -> pd.DataFrame:
        """Return pandas.DataFrame with PIN features."""
        pin_cols = ["SpecId", "Label", "ScanNr", "Peptide", "Proteins"]
        peprec_cols = [
            "spec_id",
            "peptide",
            "modifications",
            "charge",
            "observed_retention_time"
        ]
        misc_cols = ["tandem_id", "ModPeptide"]
        non_feature_cols = pin_cols + peprec_cols + misc_cols
        feature_cols = [col for col in self.df.columns if col not in non_feature_cols]
        return self.df[feature_cols]

    # TODO if USI is used extract_spectrum_index should be False
    def to_peptide_record(
        self,
        extract_spectrum_index: Optional[bool] = True,
        spectrum_index_pattern: Optional[str] = None,
        score_column_label: Optional[str] = None,
    ) -> PeptideRecord:
        """
        Convert PercolatorIn to PeptideRecord.

        Parameters
        ----------
        extract_spectrum_index: bool (optional)
            If True, extract spectrum index from spec_id `SpecId` with the
            `spectrum_index_pattern` regex pattern, or the default pattern:
            `.+_([0-9]+)_[0-9]+_[0-9]+`. If False, copy the original values in `SpecId`.
        spectrum_index_pattern: str (optional)
            Regex pattern with one group matching the MGF spectrum index in the PIN
            `SpecId` column.
        score_column_label: str (optional)
            Label of column to copy to PeptideRecord as `psm_score`. If None, common
            column labels will be tested and used if available.

        Returns
        -------
        ms2rescore.peptide_record.PeptideRecord

        """
        # Assign one of the default score column labels, if available
        if not score_column_label:
            for col in ["lnEValue", "hyperscore", "Score", "COMET:lnExpect"]:
                if col in self.df.columns:
                    score_column_label = col
                    logger.debug(
                        "Found score column in PIN with label %s", score_column_label
                    )
                    break
            if not score_column_label:
                logger.debug("No known score column found.")

        # Create new peprec
        peprec_df = pd.DataFrame()
        if extract_spectrum_index:
            peprec_df["spec_id"] = self._get_spectrum_index_column(
                spectrum_index_pattern
            )
        else:
            peprec_df["spec_id"] = self.df["SpecId"]
        peprec_df["peptide"] = self._get_sequence_column()
        peprec_df["modifications"] = self._get_peprec_modifications_column()
        peprec_df["charge"] = self._get_charge_column()
        if score_column_label:
            peprec_df["psm_score"] = self.df[score_column_label]
        peprec_df["Label"] = self.df["Label"]
        peprec_df["Proteins"] = self.df["Proteins"]
        if "RT" in self.df.columns:
            peprec_df["observed_retention_time"] = self.df["RT"]
        peprec = PeptideRecord.from_dataframe(peprec_df)

        return peprec


def run_percolator_converter(
    converter_name: str,
    path_to_id_file: str,
    path_to_pin_file: str,
    id_decoy_pattern: Optional[str] = None,
    log_level: str = "info",
):
    """
    Run Percolator PIN Converter.

    Parameters
    ----------
    converter_name: str
        name of the converter to use; currently supported are {"msgf2pin", "tandem2pin"}
    path_to_id_file: str
        path to identification file (input to converter)
    path_to_pin_file: str'
        path to pin file (output of converter)
    id_decoy_pattern: str, optional
        identifications file decoy pattern (e.g. REVERSED), passed to `-P` option of
        Percolator converter; default depends on input filetype
    log_level: str, optional
        log_level, default is `info`

    """
    default_decoy_patterns = {"msgf2pin": "XXX", "tandem2pin": "REVERSED"}
    if converter_name not in default_decoy_patterns.keys():
        raise NotImplementedError(converter_name)
    if not id_decoy_pattern:
        id_decoy_pattern = default_decoy_patterns[converter_name]

    command = [
        converter_name,
        "-P",
        id_decoy_pattern,
        "-o",
        os.path.abspath(path_to_pin_file),
        os.path.abspath(path_to_id_file),
    ]

    logger.info("Running Percolator PIN converter")
    logger.debug(' '.join(command))

    subprocess.run(command, capture_output=log_level == "debug", check=True)



"""PeptideShaker Extended PSM Report."""

import logging
import os
from typing import Union

import click
import numpy as np
import pandas as pd

from ms2rescore.peptide_record import PeptideRecord

logger = logging.getLogger(__name__)


@pd.api.extensions.register_dataframe_accessor("ext_psm_report")
class ExtendedPsmReportAccessor:
    """
    Pandas extension for PeptideShaker Extended PSM Reports.

    Examples
    --------
    >>> import pandas as pd
    >>> from ms2rescore.peptideshaker import ExtendedPsmReportAccessor
    >>> psm_report = pd.DataFrame.ext_psm_report.from_tsv(kwargs["input_psm_report"])
    >>> peprec = psm_report.ext_psm_report.to_peprec()
    """

    def __init__(self, pandas_obj: pd.DataFrame) -> None:
        """Pandas extension for PeptideShaker Extended PSM Reports."""
        self._obj = pandas_obj
        self._validate()

    def _validate(self):
        """Validate Pandas DataFrame as Extended PSM Report."""
        # TODO: Implement validation of PSM report DataFrame
        self.drop_invalid_amino_acids()

    def drop_invalid_amino_acids(self, invalid_amino_acids=r"[BJOUXZ]"):
        """Drop all PSMs (rows) with peptides containing invalid amino acids."""
        to_drop = self._obj[
            self._obj['Sequence'].str.contains(invalid_amino_acids, regex=True)
        ].index
        if len(to_drop) > 0:
            logger.warning(
                "Dropping %i PSMs from report due to invalid amino acids (%s)",
                len(to_drop),
                invalid_amino_acids
            )
            self._obj = self._obj.drop(index=to_drop)

    @staticmethod
    def from_tsv(path: Union[str, os.PathLike]) -> pd.DataFrame:
        """Read Extended PSM Report from TSV file."""
        ext_psm_report = pd.read_csv(path, sep="\t", index_col=0)
        ext_psm_report.ext_psm_report._validate()
        return ext_psm_report

    @staticmethod
    def from_xls(path: Union[str, os.PathLike]) -> pd.DataFrame:
        """Read Extended PSM Report from XLS file."""
        ext_psm_report = pd.read_excel(path, sheet_name=0, index_col=0)
        pd.ext_psm_report._validate(ext_psm_report)
        return ext_psm_report

    @staticmethod
    def from_file(path: Union[str, os.PathLike]) -> pd.DataFrame:
        """Read Extended PSM Report from file, inferring filetype from extension."""
        ext = os.path.splitext(path)[-1].lower()
        if (ext == ".tsv") or (ext == ".txt"):
            return pd.DataFrame.ext_psm_report.from_tsv(path)
        elif (ext == ".xls") or (ext == ".xlsx"):
            return pd.DataFrame.ext_psm_report.from_xls(path)
        else:
            raise NotImplementedError(
                f"Extended PSM Report with filetype extension {ext} is not supported."
            )

    @staticmethod
    def _parse_modification(modified_seq):
        """
        Parse modified sequence to peprec modification string.

        TODO: Do not hardcode modification mapping.
        TODO: Refactor method (e.g. use regex for matching).
        TODO: Parse C-term modifications

        """
        # Initiate variables for nterm, seq and cterm
        mod_list = list()
        nterm, seq, cterm = modified_seq.split("-")

        # Initiatle variable for nterm
        pyro_bool = False

        # Initiate variables for seq
        mod_index = 0
        mod_description = False  # to check if it's an amino acid (False) or a description in < ... > (True)

        # Check amino terminus for modifications
        if nterm == "ace":
            mod_list.append("0|Acetyl")
        elif nterm == "pyro":
            pyro_bool = True
        elif nterm != "NH2":
            print("Unknown N-terminal modification: {}".format(nterm))

        # Check internal sequence
        for char in seq:
            if char == "<":
                mod_peprec = "{}|".format(mod_index)
                mod_name = ""
                mod_description = True
            elif char == ">":
                mod_description = False
                if mod_name == 'ox':
                    mod_peprec += 'Oxidation'
                elif mod_name == 'cmm':
                    mod_peprec += 'Carbamidomethyl'
                elif mod_name == 'deam':
                    mod_peprec += 'Deamidated'
                else:
                    logger.warning("Unknown internal modification: %s", mod_name)
                mod_list.append("{}".format(mod_peprec))  # peprec format
                mod_peprec = ""

            else:
                if pyro_bool:
                    if char == 'C':
                        mod_name = "Pyro-carbamidomethyl"
                    elif char == 'Q':
                        mod_name = "Gln->pyro-Glu"
                    elif char == 'E':
                        mod_name = "Glu->pyro-Glu"
                    elif char == 'P':
                        mod_name = "Pro->pyro-Glu"
                    else:
                        logger.warning("Unknown N-terminal pyro modification from %s", char)
                    mod_list.append("1|{}".format(mod_name))
                    pyro_bool = False
                    mod_index += 1
                    mod_name = ""
                else:
                    if mod_description:
                        mod_name += char
                    else:
                        mod_index += 1

        mods_peprec = "|".join(mod_list)
        if mods_peprec == "":
            mods_peprec = "-"

        return mods_peprec

    def to_peprec(self):
        """Convert Extended PSM Report to PEPREC."""
        column_mapping = {
            "Spectrum Title": "spec_id",
            "Modified Sequence": "modifications",
            "Sequence": "peptide",
            "Measured Charge": "charge",
            "Decoy": "Label",
            "RT": "observed_retention_time",
            "Confidence [%]": "psm_score",
        }

        # Convert DataFrame to PEPREC
        df = self._obj[column_mapping.keys()].rename(columns=column_mapping)
        df["charge"] = df["charge"].str.strip("+")
        df["modifications"] = df["modifications"].apply(self._parse_modification)
        df["Label"] = df["Label"].apply(
            lambda x: 1 if x == 0 else (-1 if x == 1 else np.nan)
        )
        if df["Label"].isna().any():
            raise ValueError(
                "Missing target/decoy labels in PeptideShaker Extended PSM "
                "Report."
            )

        peprec = PeptideRecord()
        peprec.df = df
        return peprec

    def get_search_engine_features(self):
        """Get pandas.DataFrame with search engine features."""
        # TODO: Implement this!
        raise NotImplementedError


@click.command()
@click.argument("input-psm-report")
@click.argument("output-peprec")
def main(**kwargs):
    """Convert Extended PSM Report to PEPREC."""
    psm_report = pd.DataFrame.ext_psm_report.from_file(kwargs["input_psm_report"])
    peprec = psm_report.ext_psm_report.to_peprec()
    peprec.to_csv(kwargs["output_peprec"])


if __name__ == "__main__":
    main()



"""Plot MS²ReScore results."""

from abc import ABC
from collections import defaultdict
from typing import List, Optional

import matplotlib.axes
import matplotlib.pyplot as plt
import matplotlib.backends.backend_pdf
import numpy as np
import pandas as pd
import seaborn as sns
from pyteomics.auxiliary import qvalues
from statsmodels.distributions.empirical_distribution import ECDF


from ms2rescore.percolator import PercolatorIn
from ms2rescore._exceptions import MS2RescoreError


class RescoreRecord(ABC):
    rerecs = []
    loss_gain_df = pd.DataFrame()
    unique_df = pd.DataFrame()
    count_df = pd.DataFrame()

    def __init__(self) -> None:
        pass

    @classmethod
    def empty_rerecs(cls):
        cls.rerecs = []

    @classmethod
    def show_rerec_items(cls):
        for rerec in cls.rerecs:
            print(rerec)

    def target_decoy_distribution(
        self,
        q_label: str = "q",
        score_label: str = "score",
        decoy_label: str = "is decoy",
        score_name: str = "Score",
        fdr_threshold: Optional[float] = None,
        plot_title: str = "",
    ):
        """
        Plot target-decoy distributions.

        Plot for a given search engine output the target-decoy score distributions, the
        relation between the q-values and the PSM scores and a PP plot between the
        target and the decoy distribution.

        Parameters
        ----------
        df : pd.DataFrame
            Pandas DataFrame containing all rank 1 PSMs
        q_label : str, optional
            Label of column containing each PSM's q-value, default `q`
        score_label : str, optional
            Label of column containing the search engine scores, default `score`
        decoy_label : str, optional
            Label of column marking decoy PSMs as True and target PSMs as False, default
            `is decoy`
        score_name : str, optional
            Score name used in axis labels
        fdr_threshold : float, optional
            FDR threshold to plot as vertical dotted line
        plot_title : str, optional
            plot suptitle

        Returns
        -------
        fig : Matplotlib figure
        axes : Matplotlib Axes

        """
        fig, axes = plt.subplots(1, 3, figsize=(16, 4))

        if fdr_threshold:
            score_cutoff = (
                self.df[(self.df[q_label] <= 0.01) & (~self.df[decoy_label])]
                .sort_values(q_label)
                .iloc[-1][score_label]
            )
        else:
            score_cutoff = None

        # Score distribution plot
        plot_list = [
            list(self.df[self.df[decoy_label]][score_label]),
            list(self.df[~self.df[decoy_label]][score_label]),
        ]
        axes[0].hist(
            plot_list,
            bins=24,
            label=["Decoy", "Target"],
            color=["r", "blue"],
            lw=1,
            rwidth=1,
        )
        if fdr_threshold:
            axes[0].vlines(
                x=score_cutoff, ymin=0, ymax=axes[0].get_ylim()[1], linestyles="dashed"
            )
        axes[0].legend()
        axes[0].set_ylabel("Number of matches")
        axes[0].set_xlabel(score_name)

        # Q value plot
        axes[1].plot(
            self.df.sort_values(score_label)[score_label],
            self.df.sort_values(score_label)[q_label],
        )
        if fdr_threshold:
            axes[1].vlines(
                x=score_cutoff, ymin=0, ymax=axes[1].get_ylim()[1], linestyles="dashed"
            )
        axes[1].set_ylabel("q-value")
        axes[1].set_xlabel(score_name)

        # PP plot
        ratio = (
            self.df[decoy_label].value_counts()[True]
            / self.df[decoy_label].value_counts()[False]
        )
        Ft = ECDF(self.df[~self.df[decoy_label]][score_label])
        Fd = ECDF(self.df[self.df[decoy_label]][score_label])
        x = self.df[~self.df[decoy_label]][score_label]
        Fdp = Fd(x)
        Ftp = Ft(x)
        axes[2].scatter(Fdp, Ftp, s=4)
        axes[2].plot((0, 1), (0, ratio), color="r")
        axes[2].set_xlabel("Decoy percentile")
        axes[2].set_ylabel("Target percentile")

        plt.suptitle(plot_title)
        sns.despine()

        return fig, axes

    @classmethod
    def qvalue_comparison(
        cls,
        dataset_labels: Optional[List[str]] = None,
        q_label: str = "q",
        decoy_label: str = "is decoy",
        fdr_thresholds: Optional[List[float]] = None,
        log_scale: bool = True,
        title: str = "",
        ax: Optional[matplotlib.axes.Axes] = None,
    ):
        """
        Plot identification count in function of q-value threshold for multiple
        datasets.

        Parameters
        ----------
        datasets : List[pd.DataFrame]
            list of datasets in the form of `pandas.DataFrame`s
        dataset_labels : List[str]
            list of dataset labels to use in figure legend
        q_label : str, optional
            label of column containing each PSM's q-value, default `q`
        decoy_label : str, optional
            label of column marking decoy PSMs as True and target PSMs as False, default
            `is decoy`
        fdr_thresholds : List[float], optional
            list of FDR thresholds to plot as vertical, dotted lines
        log_scale : bool
            plot x-axis (q-values) in log scale or not.
        ax : matplotlib Axes, optional
            axes object to draw the plot onto, otherwise uses the current axes.

        Returns
        -------
        ax : matplotlib Axes


        """
        fig = plt.figure()
        if not isinstance(cls.rerecs, list):
            raise TypeError("`datasets` should be of type `list`.")
        if not cls.rerecs:
            raise ValueError("`datasets` cannot be empty.")

        if not fdr_thresholds:
            fdr_thresholds = [0.01, 0.001]

        if ax is None:
            ax = plt.gca()

        max_count = 0

        for i, rerec in enumerate(cls.rerecs):
            # Cumulatively count target IDs at each FDR threshold
            df = (
                rerec.df[~rerec.df[decoy_label]]
                .reset_index(drop=True)
                .sort_values(q_label, ascending=True)
                .copy()
            )
            df["count"] = (~df[decoy_label]).cumsum()

            # Plot counts
            if dataset_labels:
                label = dataset_labels[i]
            else:
                label = rerec.rescore_features
            ax.plot(df[q_label], df["count"], label=label, alpha=0.5)

            # Get maximum count, required for vertical line height
            tmp_max = np.max(df["count"])
            if tmp_max > max_count:
                max_count = tmp_max

        # Plot FDR thresholds as dotted lines
        if fdr_thresholds:
            for fdr in fdr_thresholds:
                ax.plot(
                    [fdr] * 2,
                    np.linspace(0, max_count, 2),
                    linestyle="--",
                    color="black",
                )

        # Figure labels and legend
        ax.set_xlim(0.00001, 1)
        ax.set_ylabel("Number of identified spectra")
        ax.set_title(title)
        if log_scale:
            ax.set_xlabel("FDR threshold (log scale)")
            ax.set_xscale("log")
        else:
            ax.set_xlabel("FDR threshold")

        ax.legend()
        fig.set_size_inches(12, 10)

        return ax

    @classmethod
    def _count_identifications(cls, FDR_threshold=[0.01]):
        """
        Count the number of identifications for each PIN, POUT that is listed
        in the rescore record for certain FDR thresholds.

        Parameters
        ----------
        FDR_threshold : List[floats]
            list of FDR threshold for which to count number of identifications

        """
        identification_counts = []
        for rerec in cls.rerecs:
            for threshold in FDR_threshold:
                tmp = []
                tmp.append(rerec.name)
                tmp.append(rerec.rescore_features)
                tmp.append(sum(rerec.df.q[~rerec.df["is decoy"]] < threshold))
                tmp.append(threshold)
                identification_counts.append(tmp)
        cls.count_df = pd.DataFrame(
            identification_counts, columns=["sample", "rescoring", "count", "FDR"]
        )

    @classmethod
    def _separate_unique_peptides(cls, FDR_threshold=[0.01]):
        """
        Count the number of unique identifications for each PIN, POUT that is listed in
        the rescore record for certain FDR thresholds.

        Parameters
        ----------
        FDR_threshold : List[floats]
            list of FDR threshold for which to count number of unique identifications

        """
        unique_samples = []
        for rerec in cls.rerecs:
            for threshold in FDR_threshold:
                tmp = []
                tmp.append(rerec.name)
                tmp.append(rerec.rescore_features)
                tmp.append(
                    rerec.df["peptide"][
                        ~(rerec.df["is decoy"]) & (rerec.df["q"] < threshold)
                    ].unique()
                )
                tmp.append(threshold)
                unique_samples.append(tmp)
        cls.unique_df = pd.DataFrame(
            unique_samples, columns=["sample", "rescoring", "upeps", "FDR"]
        )
        cls.unique_df["count"] = cls.unique_df["upeps"].apply(len)

    @classmethod
    def count_plot(cls, unique=False):
        """
        Barplot of the number of (unique) identifications for each FDR threshold.

        Parameters
        ----------
        unique : Boolean
            If true only plot unique identifications
            If false plot total amount of identifications
        """
        if unique:
            if cls.unique_df.empty:
                cls._separate_unique_peptides()
            y_label = "number of unique identified peptides"
            count_df = cls.unique_df
        else:
            if cls.count_df.empty:
                cls._count_identifications()
            y_label = "number of identified peptides"
            count_df = cls.count_df

        g = sns.catplot(
            x="sample",
            y="count",
            data=count_df,
            hue="rescoring",
            kind="bar",
            col="FDR",
            legend=False,
        )

        g.set_ylabels(y_label)
        g.set_xlabels("")
        g.add_legend(loc=7)
        g.fig.set_size_inches(12, 10)

        return g

    @classmethod
    def calculate_loss_gain_df(cls, reference=None, FDR_threshold=[0.01]):
        """
        Calculate relative amount of gained, lossed and shared unique peptides for each
        rescoring method relative to the reference.

        Parameters
        ----------
        reference : string
            The rescoring method be used as a reference

        FDR_threshold : List[floats]
            list of FDR threshold for which to count relative gain, loss

        """
        if cls.unique_df.empty:
            cls._separate_unique_peptides()
        loss_gain = defaultdict(list)
        for sample in cls.unique_df["sample"].unique():
            for threshold in FDR_threshold:
                ft_dict = {}
                for feature in cls.unique_df["rescoring"][
                    cls.unique_df["sample"] == sample
                ].unique():
                    ft_dict[feature] = set(
                        list(
                            cls.unique_df["upeps"][
                                (cls.unique_df["sample"] == sample)
                                & (cls.unique_df["rescoring"] == feature)
                                & (cls.unique_df["FDR"] == threshold)
                            ].item()
                        )
                    )
                if reference:
                    pass
                elif "After rescoring: Searchengine" in ft_dict.keys():
                    reference = "After rescoring: Searchengine"
                else:
                    reference = "Before rescoring"

                total = len(ft_dict[reference])
                if total == 0:
                    continue
                for feature in cls.unique_df["rescoring"][
                    cls.unique_df["sample"] == sample
                ].unique():
                    loss_gain["sample"].append(sample)
                    loss_gain["feature"].append(feature)
                    loss_gain["FDR"].append(threshold)
                    loss_gain["shared"].append(
                        (len(ft_dict[reference] & ft_dict[feature]) / total) * 100
                    )
                    loss_gain["gain"].append(
                        (len(ft_dict[feature] - ft_dict[reference]) / total) * 100
                    )
                    loss_gain["loss"].append(
                        (len(ft_dict[reference] - ft_dict[feature]) / total) * -100
                    )

        cls.loss_gain_df = pd.DataFrame(loss_gain)
        cls.loss_gain_df["gain"] = cls.loss_gain_df["gain"] + cls.loss_gain_df["shared"]

    @classmethod
    def loss_gain_plot(cls, FDR):
        """
        Barplot of the relative count of unique peptides for a fixed FDR
        for each rescoring feature set.

        Parameters
        ----------
        FDR : float
            single FDR threshold used for the plot
        """
        fig = plt.figure()

        if cls.loss_gain_df.empty:
            cls.calculate_loss_gain_df(FDR_threshold=[FDR])
        if FDR not in cls.loss_gain_df["FDR"].unique():
            cls._separate_unique_peptides(FDR_threshold=[FDR])
            cls.calculate_loss_gain_df(FDR_threshold=[FDR])

        tmp = cls.loss_gain_df[cls.loss_gain_df["FDR"] == FDR]

        for sample in zip(
            tmp["sample"].unique(), list(range(1, len(tmp["sample"].unique()) + 1))
        ):
            if sample[1] == len(tmp["sample"].unique()):
                ax = fig.add_subplot(
                    int(f"{len(tmp['sample'].unique())}1{sample[1]}"), frameon=False
                )
                ax.title.set_text(f"{sample[0]}\nFDR={FDR}")
                sns.barplot(
                    y="feature",
                    x="gain",
                    data=tmp[tmp["sample"] == sample[0]],
                    color="#2FA92D",
                    order=tmp[tmp["sample"] == sample[0]].sort_values("gain").feature,
                    ax=ax,
                )
                sns.barplot(
                    y="feature",
                    x="shared",
                    data=tmp[tmp["sample"] == sample[0]],
                    color="#1AA3FF",
                    order=tmp[tmp["sample"] == sample[0]].sort_values("gain").feature,
                    ax=ax,
                )
                sns.barplot(
                    y="feature",
                    x="loss",
                    data=tmp[tmp["sample"] == sample[0]],
                    color="#FF0000",
                    order=tmp[tmp["sample"] == sample[0]].sort_values("gain").feature,
                    ax=ax,
                )
                ax.axes.set_xlabel("unique identified peptides (%)")
                ax.axes.set_ylabel("")
            else:
                ax = fig.add_subplot(
                    int(f"{tmp['sample'].unique()}1{sample[1]}"), frameon=False
                )
                ax.title.set_text(sample[0])
                sns.barplot(
                    y="feature",
                    x="gain",
                    data=tmp[tmp["sample"] == sample[0]],
                    color="#2FA92D",
                    order=tmp[tmp["sample"] == sample[0]].sort_values("gain").feature,
                    ax=ax,
                )
                sns.barplot(
                    y="feature",
                    x="shared",
                    data=tmp[tmp["sample"] == sample[0]],
                    color="#1AA3FF",
                    order=tmp[tmp["sample"] == sample[0]].sort_values("gain").feature,
                    ax=ax,
                )
                sns.barplot(
                    y="feature",
                    x="loss",
                    data=tmp[tmp["sample"] == sample[0]],
                    color="#FF0000",
                    order=tmp[tmp["sample"] == sample[0]].sort_values("gain").feature,
                    ax=ax,
                )
                ax.axes.set_xlabel("")
                ax.axes.set_ylabel("")
        plt.grid(axis="x")
        ax.set_axisbelow(True)
        fig.set_size_inches(12, 6)
        return ax

    @classmethod
    def save_plots_to_pdf(cls, filename, FDR_thresholds=[0.01]):
        """
        Create pdf file with rescore plots.

        Parameters
        ----------
        filename : string
            filename for the pdf file

        FDR_thresholds : list[float]
            list of FDR thresholds for wich to create plots

        """
        if not cls.rerecs:
            raise MS2RescoreError("no pin/pout files listed")

        pdf = matplotlib.backends.backend_pdf.PdfPages(filename)
        cls._separate_unique_peptides(FDR_thresholds)
        cls._count_identifications(FDR_thresholds)

        cls.count_plot(unique=False)
        pdf.savefig()

        cls.qvalue_comparison()
        plt.tight_layout()
        pdf.savefig()

        cls.loss_gain_plot(FDR=0.01)
        plt.tight_layout()
        pdf.savefig()
        pdf.close()


class PIN(RescoreRecord):
    """PIN file record."""

    def __init__(self, path_to_file, sample_name, score_metric=None) -> None:
        super().__init__()
        self.type = "pin"
        self.rescore_features = "Before rescoring"
        self.name = sample_name
        if self._infer_filetype(path_to_file) == "pin":
            self.score_metric = score_metric
            self.df = self._read_pin_file(path_to_file)
        elif self._infer_filetype(path_to_file) == "peprec":
            self.score_metric = "psm_score"
            self.df = self._read_pin_from_peprec(path_to_file)
        else:
            raise MS2RescoreError("Not a peprec or pin file")

        self.rerecs.append(self)  # add PIN record to RescoreRecord

        """
        Parameters
        ----------
        path_to_file : string
            path to input file, either peprec or pin file

        sample_name : string
            name of the sample

        score_metric: string
            score column label (only required when providing pin file)
        """

    def __str__(self) -> str:
        return f"Sample name: {self.name}\nType: {self.type}\nRescore status: {self.rescore_features}"

    def _read_pin_file(self, path_to_pin):
        """Read pin file, calculate qvalues and write into single pandas DataFrame."""
        pin = PercolatorIn(path_to_pin)
        pin_qvalues = pd.DataFrame(
            qvalues(
                pin.df,
                key=pin.df[self.score_metric],
                is_decoy=pin.df["Label"] == -1,
                reverse=True,
                remove_decoy=False,
                formula=1,
                full_output=True,
            )
        )
        return pin_qvalues[["SpecId", "is decoy", "score", "q", "Peptide"]].rename(
            columns={"SpecId": "PSMId", "Peptide": "peptide"}
        )

    def _read_pin_from_peprec(self, path_to_peprec):
        peprec = pd.read_table(path_to_peprec, sep=" ")
        pin_qvalues = pd.DataFrame(
            qvalues(
                peprec,
                key=peprec["psm_score"],
                is_decoy=peprec["Label"] == -1,
                reverse=True,
                remove_decoy=False,
                formula=1,
                full_output=True,
            )
        )
        return pin_qvalues[["spec_id", "is decoy", "score", "q", "peptide"]].rename(
            columns={"spec_id": "PSMId"}
        )

    def _infer_filetype(self, filepath):
        """
        Extract filetype from file

        Parameters
        ----------
        filepath : string
            filepath from file

        """
        filetype = filepath.rsplit(".", 1)[1]
        return filetype


class POUT(RescoreRecord):
    """POUT file record."""

    def __init__(
        self, path_to_target_pout, path_to_decoy_pout, sample_name, rescoring_features
    ) -> None:
        super().__init__()
        self.target_pout = path_to_target_pout
        self.decoy_pout = path_to_decoy_pout
        self.df = self._read_pout_file
        self.type = "pout"
        self.rescore_features = "After rescoring: " + rescoring_features
        self.name = sample_name
        self.df = self._read_pout_file()
        self.rerecs.append(self)  # add POUT record to RescoreRecord

        """
        Parameters
        ----------
        path_to_target_pout : string
            path to target pout file

        path_to_decoy_pout : string
            path to decoy pout file

        sample_name : string
            name of the sample

        rescoring_features: list[strings]
            list of the used rescoring features
        """

    def __str__(self) -> str:
        return f"Sample name: {self.name}\nType: {self.type}\nRescore status: {self.rescore_features}"

    @staticmethod
    def _read_pout(path):
        pout = pd.read_csv(PercolatorIn.fix_tabs(path, id_column="PSMId"), sep="\t")
        return pout

    def _read_pout_file(self):
        """Read target and decoy pout files and combine into single pandas DataFrame."""

        target_pout = self._read_pout(self.target_pout)
        decoy_pout = self._read_pout(self.decoy_pout)
        target_pout["Label"] = 1
        decoy_pout["Label"] = -1
        pout = pd.concat([target_pout, decoy_pout])

        pout_qvalues = pout[["PSMId", "score", "q-value", "Label", "peptide"]].rename(
            columns={"q-value": "q", "Label": "is decoy"}
        )
        pout_qvalues["is decoy"] = pout["Label"] == -1

        return pout_qvalues



#! python
"""Setup ms2rescore."""

from setuptools import setup


def get_version(path):
    """Get __version__ from Python file."""
    with open(path, "rt") as f:
        for line in f:
            if line.startswith("__version__ = "):
                return line.strip().split(" = ")[1].strip("\"'")


def get_readme():
    with open("README.md", "r") as fh:
        readme = fh.read()
    return readme


setup(
    name="ms2rescore",
    version=get_version("ms2rescore/_version.py"),
    license="apache-2.0",
    description="MS²Rescore: Sensitive PSM rescoring with predicted MS² peak intensities and retention times.",
    long_description=get_readme(),
    long_description_content_type="text/markdown",
    author="Ralf Gabriels, Arthur Declercq, Ana Sílvia C. Silva, Tim Van Den Bossche",
    author_email="compomics.list@gmail.com",
    url="https://compomics.github.io/projects/ms2rescore/",
    project_urls={
        "Documentation": "http://compomics.github.io/projects/ms2rescore",
        "Source": "https://github.com/compomics/ms2rescore",
        "Tracker": "https://github.com/compomics/ms2rescore/issues",
        "Publication": "https://doi.org/10.1093/bioinformatics/btz383",
    },
    classifiers=[
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Development Status :: 4 - Beta",
    ],
    keywords=[
        "MS2Rescore",
        "MS2PIP",
        "DeepLC",
        "Percolator",
        "Proteomics",
        "peptide",
        "peak intensity prediction",
        "spectrum",
        "machine learning",
    ],
    packages=["ms2rescore"],
    include_package_data=True,
    entry_points={"console_scripts": [
        "ms2rescore=ms2rescore.__main__:main",
    ]},
    python_requires=">=3.7",
    install_requires=[
        "importlib-resources;python_version<'3.7'",
        "numpy>=1.16.0,<2",
        "pandas>=0.24.0,<2",
        "scikit-learn>=0.20.0,<1",
        "scipy>=1.2.0,<2",
        "tqdm>=4.31.0,<5",
        "pyteomics>=4.1.0,<5",
        "lxml>=4.5,<5",
        "ms2pip>=3.8,<4",
        "click>=7,<8",
        "cascade-config>=0.3.0,<2",
        "matplotlib>=3,<4",
        "seaborn>=0.11",
        "statsmodels>=0.12",
        "deeplc>=0.1.17",
    ],
    extras_require={
        "gui": ["gooey>=1.0"],
    },
    test_suite="tests",
    tests_require=["ms2rescore", "pytest>=4.3.0,<5"],
)
